/* compiler.pl */
/* This file contains the predicates for compiling a DEVICE
   production rule into an EXACT ECA rule, using a complex
   event network for incremental matching of conditions 
   at run-time
*/

/* This predicate compiles a DEVICE rule */
compile(Rule,AuxAtts,RuleOID) :- 
	parse(Rule,Tree),
	compile_rule(Tree,AuxAtts,RuleOID).

/* This predicate compiles a DEVICE rule and resturns the Rule OID of the created ECA rule */
compile_rule([CondIn,Action],AuxAtts,RuleOID) :- 
	(append(Cond,[prolog(Goal)],CondIn) -> 
		true;
		Cond = CondIn),
	first_pass(1,Cond,OID_Var_list1,[],Var_list1,Top_event_list1,Negative_list1,[],UniqueList,No_of_events1,Var_groups),
	check_vars_in_action(Action,[OID_Var_list1,Var_list1],
				    [OID_Var_list1_1,Var_list1_1],Cond2ActionVars1),
	(nonvar(Goal) -> 
		(check_vars_in_action(Goal,[OID_Var_list1_1,Var_list1_1],
					   [OID_Var_list2,Var_list2],Cond2ActionVars2),
		 union(Cond2ActionVars1,Cond2ActionVars2,Cond2ActionVars));
		(OID_Var_list2 = OID_Var_list1_1,
		 Var_list2 = Var_list1_1,
		 Cond2ActionVars = Cond2ActionVars1,
		 Goal = true)),
	fix_list(Var_list2,Var_list3),
	second_pass_stage1(OID_Var_list2,Var_list3,Top_event_list1,Top_event_list2,
				Negative_list1,Negative_list2,UniqueList,No_of_events2,Var_groups),
	second_pass_stage2(Top_event_list2,Negative_list2,No_of_events3,Last_event_OID),
	No_of_events is No_of_events1 + No_of_events2 + No_of_events3,
	create_rule(Last_event_OID,Goal,Action,AuxAtts,No_of_events,Cond2ActionVars,RuleOID).


/* first_pass(Cond,OID_Var_list1,Var_list1,Top_event_list,Negative_list,No_of_primitive_events) */
first_pass(_,[],[],Var_list,Var_list,[],[],Unique,Unique,0,[]) :- !.
first_pass(I,[nested_neg(IOPs)|RestCond],[Nested_OID_Var_list|Rest_OID_Var_list],Prev_Var_list,Var_list,
	   Top_event_list,[var(I,Nested_Negative_events,0)|Rest_Neg],PrevUnique,FinalUnique,Number,Var_groups) :- !,
	first_pass(1,IOPs,Nested_OID_Var_list,Prev_Var_list,Next_Var_list,Nested_Top_event_list,
			_Nested_Negative_list,PrevUnique,PrevUnique,NestedNumber,Nested_Var_groups),
	extract_negative_events(Nested_OID_Var_list,Nested_Negative_events),
	I1 is I + 1,
	first_pass(I1,RestCond,Rest_OID_Var_list,Next_Var_list,Var_list,Rest_Top,
			Rest_Neg,PrevUnique,FinalUnique,PreviousNumber,Rest_var_groups),
	append(Nested_Var_groups,Rest_var_groups,Var_groups),
	append(Nested_Top_event_list,Rest_Top,Top_event_list),
	Number is NestedNumber + PreviousNumber.
first_pass(I,[Cond|RestCond],[var(OID_Var,Event_list1,0)|Rest_OID],Var_list,Var_list,
	   Top_event_list,[var(I,NegEvents,0)|Rest_Neg],PrevUniqueList,FinalUniqueList,
		Number,[var(OID_Var,Var_group)|Rest_var_groups]) :- 
	Cond =.. [PosNeg,[OID_Var,Class,[]]], !,
	pass_class_pattern(PosNeg,OID_Var,Class,Event_list1,Top_events,NegEvents,PreviousNumber1,Var_group),
	I1 is I + 1,
	first_pass(I1,RestCond,Rest_OID,_Next_Var_list,Var_list,Rest_Top,
			Rest_Neg,PrevUniqueList,FinalUniqueList,PreviousNumber2,Rest_var_groups),
	append(Top_events,Rest_Top,Top_event_list),
	Number is PreviousNumber1 + PreviousNumber2.
% This rule handles the case: X@class(....)\=Y
first_pass(I,[Cond|RestCond],[var(OID_Var,Event_list1,0)|Rest_OID],Prev_Var_list,Var_list,
	   Top_event_list,[var(I,NegEvents,0)|Rest_Neg],PrevUniqueList,FinalUniqueList,
		Number,[var(OID_Var,Var_group)|Rest_var_groups]) :- 
	Cond =.. [PosNeg,[OID_Var\==OID_Var2,Class,AttribPatterns]], !,
	pass_attrib_patterns(PosNeg,OID_Var,Class,AttribPatterns,Event_list1,Prev_Var_list,Next_Var_list1,
	                            Top_events,NegEvents,PrevUniqueList,NextUniqueList,PreviousNumber1,Var_group),
	delete(Next_Var_list1,var(OID_Var2,OldVarList,0),Next_Var_list2),
	(var(OldVarList) -> OldVarList = []; true),
	% The entry that is kept in the NormVarList is weird, in order to be easily identified by second pass
	Next_Var_list = [var(OID_Var2,[ev([],[],['\\==',OID_Var],[])|OldVarList],0)|Next_Var_list2],
	I1 is I + 1,
	first_pass(I1,RestCond,Rest_OID,Next_Var_list,Var_list,Rest_Top,
			Rest_Neg,NextUniqueList,FinalUniqueList,PreviousNumber2,Rest_var_groups),
	append(Top_events,Rest_Top,Top_event_list),
	Number is PreviousNumber1 + PreviousNumber2.
first_pass(I,[Cond|RestCond],[var(OID_Var,Event_list1,0)|Rest_OID],Prev_Var_list,Var_list,
	   Top_event_list,[var(I,NegEvents,0)|Rest_Neg],PrevUniqueList,FinalUniqueList,
		Number,[var(OID_Var,Var_group)|Rest_var_groups]) :- 
	Cond =.. [PosNeg,[OID_Var,Class,AttribPatterns]],
	pass_attrib_patterns(PosNeg,OID_Var,Class,AttribPatterns,Event_list1,Prev_Var_list,Next_Var_list,
	                            Top_events,NegEvents,PrevUniqueList,NextUniqueList,PreviousNumber1,Var_group),
	I1 is I + 1,
	first_pass(I1,RestCond,Rest_OID,Next_Var_list,Var_list,Rest_Top,
			Rest_Neg,NextUniqueList,FinalUniqueList,PreviousNumber2,Rest_var_groups),
	append(Top_events,Rest_Top,Top_event_list),
	Number is PreviousNumber1 + PreviousNumber2.


pass_attrib_patterns(_,_,_,[],[],Var_list,Var_list,[],[],Unique,Unique,0,[]) :- !.
% This clause catters for the 'Attribute:Var' case
pass_attrib_patterns(PosNeg,OID_Var,Class,[def(Attr,var(V))|RestPatterns],
		     [ev(EventOID,AntiEventOID,['='],Out_token)|Rest_Event_list],
		     Prev_Var_list,Next_Var_list,[ev(EventOID,AntiEventOID,Out_token)|Rest_Top_events],
		     NegEvents,PrevUnique,FinalUnique,Number,[Var|RestVarGroup]) :- !,
	(V = unique(Var) ->
		NextUnique = [[Var,EventOID]|PrevUnique];
		(Var = V,
		 NextUnique = PrevUnique)),
	create_primitive_event("put_",Attr,Class,[OID_Var,Var],EventOID,Out_token),
	create_primitive_event("delete_",Attr,Class,[OID_Var,Var],AntiEventOID,Out_token),
	create_update_event(Attr,Class,Out_token,EventOID,AntiEventOID),
	% This line here checks whether the variable has been declared again, catering for the case
	% where two ':' for the same var have been given (e.g. deductive rules). If this happens
	% the second time is dealt with as a comparison (see the last case for this predicate)	
	(member(var(Var,OldVarList,0),Prev_Var_list) -> 
		(delete(Prev_Var_list,var(Var,OldVarList,0),Prev_Var_list1),
		 add_negative_info(PosNeg,EventOID,NegList),
		 Temp_var_list = [var(Var,[ev(EventOID,AntiEventOID,['='],Out_token)|OldVarList],0)|Prev_Var_list1]);
		Temp_var_list = [var(Var,[ev(EventOID,AntiEventOID,['='],Out_token)],0)|Prev_Var_list]),
	pass_attrib_patterns(PosNeg,OID_Var,Class,RestPatterns,Rest_Event_list,Temp_var_list,
			     Next_Var_list,Rest_Top_events,RestNegEvents,
				NextUnique,FinalUnique,Previous,RestVarGroup),
	(var(NegList) -> 
		NegEvents = RestNegEvents;
		append(NegList,RestNegEvents,NegEvents)),
	Number is Previous + 3.
% This clause catters for the 'Attribute Relational-Operator Constant' case
pass_attrib_patterns(PosNeg,OID_Var,Class,[comp(Attr,Preds)|RestPatterns],
		     [ev(LEventOID,LEventOID,['='],LOut_token)|Rest_Event_list],
		     Prev_Var_list,Next_Var_list,[ev(LEventOID,LEventOID,LOut_token)|Rest_Top_events],
		     NegEvents,PrevUnique,NextUnique,Number,RestVarGroup) :- 
	only_constants(Preds), !,
	create_primitive_event("put_",Attr,Class,[OID_Var,Attr],PEventOID,Out_token),
	create_primitive_event("delete_",Attr,Class,[OID_Var,Attr],AntiPEventOID,Out_token),
	create_update_event(Attr,Class,Out_token,PEventOID,AntiPEventOID),
	create_logical_event(0,OID_Var,Attr,Out_token,PEventOID,AntiPEventOID,Preds,LEventOID,LOut_token),
	add_negative_info(PosNeg,LEventOID,NegList),
	pass_attrib_patterns(PosNeg,OID_Var,Class,RestPatterns,Rest_Event_list,Prev_Var_list,Next_Var_list,
				Rest_Top_events,RestNegEvents,PrevUnique,NextUnique,Previous,RestVarGroup),
	append(NegList,RestNegEvents,NegEvents),
	Number is Previous + 4.
% This clause catters for the 'Attribute:Var Relational-Operator Constant' case
pass_attrib_patterns(PosNeg,OID_Var,Class,[defcomp(Attr,var(Var),Preds)|RestPatterns],
                        [ev(LEventOID,LEventOID,['='],LOut_token)|Rest_Event_list],
			Prev_Var_list,Next_Var_list,[ev(LEventOID,LEventOID,LOut_token)|Rest_Top_events],
                        NegEvents,PrevUnique,NextUnique,Number,[Var|RestVarGroup]) :- 
	only_constants(Preds), !,
	create_primitive_event("put_",Attr,Class,[OID_Var,Var],PEventOID,Out_token),
	create_primitive_event("delete_",Attr,Class,[OID_Var,Var],AntiPEventOID,Out_token),
	create_update_event(Attr,Class,Out_token,PEventOID,AntiPEventOID),
	create_logical_event(1,OID_Var,Var,Out_token,PEventOID,AntiPEventOID,Preds,LEventOID,LOut_token),
	pass_attrib_patterns(PosNeg,OID_Var,Class,RestPatterns,Rest_Event_list,
			     [var(Var,[ev(LEventOID,LEventOID,['='],LOut_token)],0)|Prev_Var_list],
                             Next_Var_list,Rest_Top_events,NegEvents,PrevUnique,NextUnique,Previous,RestVarGroup),
	Number is Previous + 4.
% This clause catters for the 'Attribute:Var1 Relational-Operator Var2' case
pass_attrib_patterns(PosNeg,OID_Var,Class,[defcomp(Attr,var(Var1),pred(R,var(Var2)))|RestPatterns],
	                    [ev(PEventOID,AntiPEventOID,['='],Out_token)|Rest_Event_list],
			    Prev_Var_list,Next_Var_list,[ev(PEventOID,AntiPEventOID,Out_token)|Rest_Top_events],
                            NegEvents,PrevUnique,NextUnique,Number,[Var1,Var2|RestVarGroup]) :- !,
	create_primitive_event("put_",Attr,Class,[OID_Var,Var1,Var2],PEventOID,Out_token),
	create_primitive_event("delete_",Attr,Class,[OID_Var,Var1,Var2],AntiPEventOID,Out_token),
	create_update_event(Attr,Class,Out_token,PEventOID,AntiPEventOID),
	delete(Prev_Var_list,var(Var2,OldVar2List,0),Prev_Var_list1),
	pass_attrib_patterns(PosNeg,OID_Var,Class,RestPatterns,Rest_Event_list,
			     [var(Var1,[ev(PEventOID,AntiPEventOID,['='],Out_token)],0),
                              var(Var2,[ev(PEventOID,AntiPEventOID,[R],Out_token)|OldVar2List],0)|Prev_Var_list1],
			     Next_Var_list,Rest_Top_events,NegEvents,PrevUnique,NextUnique,Previous,RestVarGroup),
	Number is Previous + 3.
% This clause catters for the 'Attribute Relational-Operator Var' case
pass_attrib_patterns(PosNeg,OID_Var,Class,[comp(Attr,pred(R,var(Var)))|RestPatterns],
                        [ev(PEventOID,AntiPEventOID,['='],Out_token)|Rest_Event_list],
			Prev_Var_list,Next_Var_list,[ev(PEventOID,AntiPEventOID,Out_token)|Rest_Top_events],
                        NegEvents,PrevUnique,NextUnique,Number,[Var|RestVarGroup]) :- !,
	create_primitive_event("put_",Attr,Class,[OID_Var,Var],PEventOID,Out_token),
	create_primitive_event("delete_",Attr,Class,[OID_Var,Var],AntiPEventOID,Out_token),
	create_update_event(Attr,Class,Out_token,PEventOID,AntiPEventOID),
	delete(Prev_Var_list,var(Var,OldVarList,0),Prev_Var_list1),
	(var(OldVarList) -> OldVarList = []; true),
	add_negative_info(PosNeg,PEventOID,NegList),
	pass_attrib_patterns(PosNeg,OID_Var,Class,RestPatterns,Rest_Event_list,
			     [var(Var,[ev(PEventOID,AntiPEventOID,[R],Out_token)|OldVarList],0)|Prev_Var_list1],
			     Next_Var_list,Rest_Top_events,RestNegEvents,PrevUnique,NextUnique,Previous,RestVarGroup),
	append(NegList,RestNegEvents,NegEvents),
	Number is Previous + 3.


pass_class_pattern(PosNeg,OID_Var,Class,[ev(EventOID,AntiEventOID,['='],Out_token)],
		   [ev(EventOID,AntiEventOID,Out_token)],NegList,2,[]) :- 
	create_class_event(new,Class,[OID_Var],EventOID,Out_token),
	create_class_event(delete,Class,[OID_Var],AntiEventOID,Out_token),
	add_negative_info(PosNeg,EventOID,NegList).


only_constants(Preds) :- 
	functor(Preds,F,_),
	memberchk(F,[single,and,or]).

add_negative_info(pos,_,[]) :- !.
add_negative_info(neg,EventOID,[EventOID]).


/* This predicate checks whether the Vars in List_in appear in the Action
   of the production rule.
   Furthermore, this predicate restores the natural order of vars and events!
*/
check_vars_in_action(Action,List_in,List_out,Cond2ActionVars) :- 
	var_list(Action,ActionVars),
	check_vars_in_action_aux(ActionVars,List_in,List_out,[],Cond2ActionVars).

check_vars_in_action_aux(_,[],[],Cond2ActionVars,Cond2ActionVars) :- !.
check_vars_in_action_aux(ActionVars,[Hin|Tin],[Hout|Tout],Temp,Cond2ActionVars) :- 
	check_vars_in_action_aux2(ActionVars,Hin,Hout,[],SomeVars),
	append(SomeVars,Temp,Next),
	check_vars_in_action_aux(ActionVars,Tin,Tout,Next,Cond2ActionVars).

check_vars_in_action_aux2(_,[],[],SomeVars,SomeVars) :- !.
check_vars_in_action_aux2(ActionVars,[var(Var,Event_list,0)|Tin],[var(Var,Event_list,1)|Tout],
	                             Temp,SomeVars) :- 
	memberchk(Var,ActionVars), !,
	check_vars_in_action_aux2(ActionVars,Tin,Tout,[var(Var,_)|Temp],SomeVars).
check_vars_in_action_aux2(ActionVars,[H|Tin],[H|Tout],Temp,SomeVars) :- 
	check_vars_in_action_aux2(ActionVars,Tin,Tout,Temp,SomeVars).



second_pass_stage1(OID_Var_listIn,Var_listIn,Top_event_listIn,Top_event_listOut,
			Negative_listIn,Negative_listOut,UniqueList,No_of_events,VarGroups) :- 
	second_pass_stage1_OID_vars(1,OID_Var_listIn,OID_Var_list1,Var_listIn,Var_list1,
	                             Top_event_listIn,Top_event_list1,Negative_listIn,Negative_list1,
				     UniqueList,UniqueList1,No_of_events1,VarGroups),
	second_pass_nested_negs(Var_list1,Var_list2,OID_Var_list1,OID_Var_list2,Top_event_list1,Top_event_list2,
                                     Negative_list1,Negative_list2,No_of_events2,VarGroups),
	re_order_vars(Var_list2,OID_Var_list2,Var_list3),
	second_pass_stage1_norm_vars(Var_list3,OID_Var_list2,Top_event_list2,Top_event_listOut,
                                     Negative_list2,Negative_listOut,UniqueList1,No_of_events3,VarGroups),
	No_of_events is No_of_events1 + No_of_events2 + No_of_events3.

re_order_vars(VarList1,OIDVarList,VarList2) :- 
	split_vars(VarList1,OIDVarList,L1,L2),
	append(L1,L2,VarList2).

split_vars([],_,[],[]) :- !.
split_vars([var(Var,EVList,Action)|T],OIDVarList,[var(Var,EVList,Action)|T1],L2) :-
	member(var(Var,_,_),OIDVarList), !,
	split_vars(T,OIDVarList,T1,L2).
split_vars([var(Var,EVList,Action)|T],OIDVarList,L1,[var(Var,EVList,Action)|T2]) :-
	split_vars(T,OIDVarList,L1,T2).

second_pass_stage1_OID_vars(_,[],[],Var_list,Var_list,Top_event_list,Top_event_list,
				Negative_list,Negative_list,Unique,Unique,0,_) :- !.
second_pass_stage1_OID_vars(I,[[var(Var,Event_list,Action)|Nested_vars]|Tail_Var_list],
	                      [var(Var,Event_listOut,Action)|Tail_Var_listOut],Var_listIn,Var_listOut,
	                     Top_event_listIn,Top_event_listOut,Negative_listIn,Negative_listOut,
		             UniqueIn,UniqueOut,No_of_events,VarGroups) :- !,
	second_pass_stage1_OID_var(I,Var,Action,Event_list,Event_listOut,Var_listIn,Var_listNext,
				   Top_event_listIn,Top_event_listNext,Negative_listIn,Negative_listNext,
				   UniqueIn,UniqueIn,No_of_events1,VarGroups),
	(Nested_vars == [] -> 
	    (I1 is I + 1,
	     second_pass_stage1_OID_vars(I1,Tail_Var_list,Tail_Var_listOut,Var_listNext,Var_listOut,
	                                    Top_event_listNext,Top_event_listOut,Negative_listNext,Negative_listOut,
					    UniqueIn,UniqueOut,No_of_events2,VarGroups));
             second_pass_stage1_OID_vars(I,[Nested_vars|Tail_Var_list],Tail_Var_listOut,Var_listNext,Var_listOut,
				           Top_event_listNext,Top_event_listOut,Negative_listNext,Negative_listOut,
					   UniqueIn,UniqueOut,No_of_events2,VarGroups)),
	No_of_events is No_of_events1 + No_of_events2.
second_pass_stage1_OID_vars(I,[var(Var,Event_list,Action)|Tail_Var_list],[var(Var,Event_listOut,Action)|Tail_Var_listOut],
	                     Var_listIn,Var_listOut,Top_event_listIn,Top_event_listOut,
			     Negative_listIn,Negative_listOut,UniqueIn,UniqueOut,No_of_events,VarGroups) :- 
	second_pass_stage1_OID_var(I,Var,Action,Event_list,Event_listOut,Var_listIn,Var_listNext,
				   Top_event_listIn,Top_event_listNext,Negative_listIn,Negative_listNext,
				   UniqueIn,UniqueNext,No_of_events1,VarGroups),
	I1 is I + 1,
	second_pass_stage1_OID_vars(I1,Tail_Var_list,Tail_Var_listOut,Var_listNext,Var_listOut,
				    Top_event_listNext,Top_event_listOut,Negative_listNext,Negative_listOut,
				    UniqueNext,UniqueOut,No_of_events2,VarGroups),
	No_of_events is No_of_events1 + No_of_events2.


second_pass_stage1_OID_var(_,_Var,_Action,[ev(EV,AEV,OP,Out_token)],[ev(EV,AEV,OP,Out_token)],Var_list,Var_list,
				Top_event_list,Top_event_list,Negative_list,Negative_list,Unique,Unique,0,_) :- !.
second_pass_stage1_OID_var(I,Var,Action,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)|RestEvents],
	                         Event_list_Out,Var_listIn,Var_listOut,Top_event_listIn,Top_event_listOut,     
				 Negative_listIn,Negative_listOut,UniqueIn,UniqueOut,No_of_events,VarGroups) :- 
	get_event_type(positive_event,PosEvent),
	create_join_event(PosEvent,Var,EV1,AEV1,OP1,Left_in_token,
			  EV2,AEV2,OP2,Right_in_token,CEventOID,['='],Out_token),
	member(var(Var,VarGroup),VarGroups),
	replace_event(VarGroup,Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                  ev(CEventOID,CEventOID,['='],Out_token),Var_listNext),
	replace_unique(VarGroup,UniqueIn,[EV1,EV2],CEventOID,UniqueNext),
	replace_negative_list(PosEvent,I,Negative_listIn,[EV1,EV2],CEventOID,Negative_listNext),
	replace_top_event(Top_event_listIn,[ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)],
		 		ev(CEventOID,CEventOID,Out_token),Top_event_listNext),
	second_pass_stage1_OID_var(I,Var,Action,[ev(CEventOID,CEventOID,['='],Out_token)|RestEvents],
	                             Event_list_Out,Var_listNext,Var_listOut,
				     Top_event_listNext,Top_event_listOut,
				     Negative_listNext,Negative_listOut,UniqueNext,UniqueOut,PrevNo_of_events,VarGroups),
	No_of_events is PrevNo_of_events + 1.


%second_pass_nested_negs(Var_list1,Var_list2,OID_Var_list1,OID_Var_list2,Top_event_list1,Top_event_list2,
%                                     Negative_list1,Negative_list2,VarGroups),
second_pass_nested_negs(Var_list,Var_list,OID_Var_list,OID_Var_list,Top_event_list,Top_event_list,[],[],0,_) :- !.
second_pass_nested_negs(Var_listIn,Var_listOut,OID_Var_listIn,OID_Var_listOut,Top_event_listIn,Top_event_listOut,
                        [var(I,NegListIn,0)|RestNegIn],[var(I,NegListOut,0)|RestNegOut],No_of_events,VarGroups) :- 
	second_pass_nested_neg(Var_listIn,Var_listNext,OID_Var_listIn,OID_Var_listNext,Top_event_listIn,Top_event_listNext,
                               NegListIn,NegListOut,No_of_events1,VarGroups),
	second_pass_nested_negs(Var_listNext,Var_listOut,OID_Var_listNext,OID_Var_listOut,Top_event_listNext,Top_event_listOut,
                                RestNegIn,RestNegOut,No_of_events2,VarGroups),
	No_of_events is No_of_events1 + No_of_events2.

second_pass_nested_neg(Var_list,Var_list,OID_Var_list,OID_Var_list,Top_event_list,Top_event_list,[],[],0,_) :- !.
second_pass_nested_neg(Var_list,Var_list,OID_Var_list,OID_Var_list,Top_event_list,Top_event_list,
                       [Last_event],[Last_event],0,_) :- !.
second_pass_nested_neg(Var_listIn,Var_listOut,OID_Var_listIn,OID_Var_listOut,Top_event_listIn,Top_event_listOut,
                               [EV1,EV2|Rest],NegListOut,No_of_events,VarGroups) :- 
	join_on_which_variable(EV1,EV2,AEV1,AEV2,Var_listIn,OID_Var_listIn,Var,OP1,Left_in_token,OP2,Right_in_token),
	get_event_type(positive_event,PosEvent),
	(var(Var) -> 
		create_cartesian_event(PosEvent,EV1,AEV1,Left_in_token,EV2,AEV2,Right_in_token,CEventOID,Out_token);
		create_join_event(PosEvent,Var,EV1,AEV1,OP1,Left_in_token,EV2,AEV2,OP2,
				  Right_in_token,CEventOID,Next_op,Out_token)),	
	replace_event(_,Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                     ev(CEventOID,CEventOID,Next_op,Out_token),Var_listNext),
	replace_event(_,OID_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                         ev(CEventOID,CEventOID,['='],Out_token),OID_Var_listNext),
	replace_top_event(Top_event_listIn,[ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)],
				ev(CEventOID,CEventOID,Out_token),Top_event_listNext),
	second_pass_nested_neg(Var_listNext,Var_listOut,OID_Var_listNext,OID_Var_listOut,
			       Top_event_listNext,Top_event_listOut,
                               [CEventOID|Rest],NegListOut,PrevNo_of_events,VarGroups),
	No_of_events is PrevNo_of_events + 1.
		

second_pass_stage1_norm_vars([],_,Top_event_list,Top_event_list,Negative_list,Negative_list,_,0,_) :- !.
second_pass_stage1_norm_vars([var(Var,Event_list,Action)|Tail_Var_listIn],OID_Var_listIn,
			     Top_event_listIn,Top_event_listOut,Negative_listIn,Negative_listOut,
			     UniqueList,No_of_events,VarGroups) :- 
	second_pass_stage1_norm_var(Var,Action,Event_list,OID_Var_listIn,OID_Var_listOut,
					Tail_Var_listIn,Tail_Var_listOut,Top_event_listIn,Top_event_listNext,
					Negative_listIn,Negative_listNext,UniqueList,UniqueListNext,No_of_events1,VarGroups),
	second_pass_stage1_norm_vars(Tail_Var_listOut,OID_Var_listOut,Top_event_listNext,Top_event_listOut,
				     Negative_listNext,Negative_listOut,UniqueListNext,No_of_events2,VarGroups),
	No_of_events is No_of_events1 + No_of_events2.



% This rule is for the case:    X@class(....)\=Y
second_pass_stage1_norm_var(Var,_Action,[ev([],[],['\\==',OtherVar],[])],OID_var_listIn,OID_Var_listOut,
	                        Tail_Var_listIn,Tail_Var_listOut,Top_event_listIn,Top_event_listOut,
				Negative_listIn,Negative_listOut,UniqueListIn,UniqueListOut,1,VarGroups) :- 
	% Check if the OID var (Y in the above example) event can be joined with another OID var (X in the above example)
	member(var(Var,[ev(EV1,AEV1,OP1,Left_in_token)],_),OID_var_listIn),
	member(var(OtherVar,[ev(EV2,AEV2,OP2,Right_in_token)],_),OID_var_listIn),
	find_related_vars(Var,VarGroups,ListGroup),
	((EV2 == EV1, AEV2 == AEV1, OP2 == OP1, Right_in_token = Left_in_token) ->
		go_back(Var,OtherVar,[EV1],[AEV1],Left_in_token,Right_in_token,Next_op,CEventOID,Out_token);
		% the following lines correct the right in-token of the join event by adding the Y variable
		(create_join_event(mm_positive_event,Var,EV1,AEV1,['='],Left_in_token,EV2,AEV2,['\\=='],
				   Right_in_token,CEventOID,Next_op,Out_token),
		 get_right_in_token(token(RL1,RL2,Right_in_templ)) => CEventOID,
		 member(EV2,RL1), member(AEV2,RL2),           		% Get the correct right in-token template
		 get_left_in_token(token(LL1,LL2,Left_in_templ)) => CEventOID,
		 member(EV1,LL1), member(AEV1,LL2),			% Get the correct left in-token template
		 member(var(Var,Val,true,true),Left_in_templ),   	% Get the Y var from left
		 member(var(OtherVar,Val,true,true),Right_in_templ),  		% Unify it with the X var on the right
		 update_right_in_token([token(RL1,RL2,Right_in_templ),     	% Store the Y var on the right 
				token(RL1,RL2,[var(Var,Val,true,true)|Right_in_templ])]) => CEventOID)),
	((EV2 == EV1, AEV2 == AEV1, CEventOID \== EV1) ->
		(replace_event(ListGroup,Tail_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                        ev(EV1,AEV1,OP1,Left_in_token),Tail_Var_listOut),
		 replace_event(_,OID_var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                               ev(EV1,AEV1,OP1,Left_in_token),OID_Var_listOut),
		 replace_negative_list(mm_positive_event,_,Negative_listIn,[EV1,EV2],EV1,Negative_listOut),
		 Top_event_listOut = Top_event_listIn);
		(replace_event(ListGroup,Tail_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                        ev(CEventOID,CEventOID,Next_op,Out_token),Tail_Var_listOut),
		 replace_event(_,OID_var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                               ev(CEventOID,CEventOID,['='],Out_token),OID_Var_listOut),
		 replace_negative_list(mm_positive_event,_,Negative_listIn,[EV1,EV2],CEventOID,Negative_listOut),
		 my_subtract(Top_event_listIn,[ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)],Top_event_listNext),
		 Top_event_listOut = [ev(CEventOID,CEventOID,Out_token)|Top_event_listNext])),
	replace_unique(ListGroup,UniqueListIn,[EV1,EV2],CEventOID,UniqueListOut).
/* The following will try to join an OID var that is also a norm var, first with another OID var and then with another norm var.
   In this way the join will become more efficient.
*/
second_pass_stage1_norm_var(Var,Action,[ev(EV1,AEV1,OP1,Left_in_token)|RestEvents],OID_var_listIn,OID_Var_listOut,
	                        Tail_Var_listIn,Tail_Var_listOut,
				Top_event_listIn,Top_event_listOut,
				Negative_listIn,Negative_listOut,UniqueListIn,UniqueListOut,No_of_events,VarGroups) :- 
	% Check if the last event can be joined with an OID var event
	nth0(Index,OID_var_listIn,var(Var,[ev(EV2,AEV2,OP2,Right_in_token)],Action)),
	RealIndex is Index + 1,
	\+ (EV2 == EV1, AEV2 == AEV1, OP2 == OP1, Right_in_token = Left_in_token), % The events should not be the same
	event_type(EV2,RealIndex,Negative_listIn,Negative_listNext1,Event_type),
	% NICK:   The following case is for a nested negative variable
	((EV2 = _#mm_negative_event, Action == 0) -> fail; !),
	create_join_event(Event_type,Var,EV1,AEV1,OP1,Left_in_token,EV2,AEV2,OP2,
				Right_in_token,CEventOID,Next_op,Out_token),
	find_related_vars(Var,VarGroups,ListGroup),
	replace_unique(ListGroup,UniqueListIn,[EV1,EV2],CEventOID,UniqueListNext),
	replace_event(ListGroup,Tail_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                        ev(CEventOID,CEventOID,Next_op,Out_token),Tail_Var_listNext),
	replace_event(_,OID_var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                               ev(CEventOID,CEventOID,['='],Out_token),OID_Var_listNext),
	replace_negative_list(Event_type,RealIndex,Negative_listNext1,[EV1,EV2],CEventOID,Negative_listNext2),
	replace_top_event(Top_event_listIn,[ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)],
					ev(CEventOID,CEventOID,Out_token),Top_event_listNext),
	second_pass_stage1_norm_var(Var,Action,[ev(CEventOID,CEventOID,Next_op,Out_token)|RestEvents],
	                        OID_Var_listNext,OID_Var_listOut,Tail_Var_listNext,Tail_Var_listOut,
				Top_event_listNext,Top_event_listOut,Negative_listNext2,Negative_listOut,
				UniqueListNext,UniqueListOut,PrevNo_of_events,VarGroups),
	No_of_events is PrevNo_of_events + 1.


second_pass_stage1_norm_var(_,_,[_],OID_var_list,OID_var_list,Tail_Var_list,Tail_Var_list,Top_event_list,Top_event_list,
				Negative_list,Negative_list,UniqueList,UniqueList,0,_) :- !.    % If no finish
second_pass_stage1_norm_var(Var,Action,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)|RestEvents],
	                        OID_Var_listIn,OID_Var_listOut,Tail_Var_listIn,Tail_Var_listOut,
				Top_event_listIn,Top_event_listOut,Negative_listIn,Negative_listOut,
				UniqueListIn,UniqueListOut,No_of_events,VarGroups) :- 
	event_type(EV2,_,Negative_listIn,Negative_listNext1,Event_type),
	create_join_event(Event_type,Var,EV1,AEV1,OP1,Left_in_token,
			  EV2,AEV2,OP2,Right_in_token,CEventOID,Next_op,Out_token),
	((member([Var,UniqueEvent],UniqueListIn),
	  member(UniqueEvent,[EV1,EV2])) ->
		put_unique_output([yes]) => CEventOID;
		true),
	find_related_vars(Var,VarGroups,ListGroup),
	replace_unique(ListGroup,UniqueListIn,[EV1,EV2],CEventOID,UniqueListNext),
	replace_event(ListGroup,Tail_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                                        ev(CEventOID,CEventOID,Next_op,Out_token),Tail_Var_listNext),
	replace_event(_,OID_Var_listIn,[ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],
	                               ev(CEventOID,CEventOID,['='],Out_token),OID_Var_listNext),
	replace_negative_list(Event_type,_,Negative_listNext1,[EV1,EV2],CEventOID,Negative_listNext2), % There are no Rest var lists
	((EV2 == EV1, AEV2 == AEV1, OP2 == OP1, Right_in_token = Left_in_token) ->
		replace_top_event(Top_event_listIn,[ev(EV1,AEV1,Left_in_token)],
					ev(CEventOID,CEventOID,Out_token),Top_event_listNext);
		replace_top_event(Top_event_listIn,[ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)],
					ev(CEventOID,CEventOID,Out_token),Top_event_listNext)),
	second_pass_stage1_norm_var(Var,Action,[ev(CEventOID,CEventOID,Next_op,Out_token)|RestEvents],
	                        OID_Var_listNext,OID_Var_listOut,Tail_Var_listNext,Tail_Var_listOut,
				Top_event_listNext,Top_event_listOut,Negative_listNext2,Negative_listOut,
				UniqueListNext,UniqueListOut,PrevNo_of_events,VarGroups),
	No_of_events is PrevNo_of_events + 1.

find_related_vars(Var,VarGroups,ListGroup) :- 
	findall(X,(member(var(X,VarGroup),VarGroups),member(Var,VarGroup)),ListGroup1),
	(member(var(Var,OwnGroup),VarGroups) -> 
		append([Var|ListGroup1],OwnGroup,ListGroup2); 
		ListGroup2 = [Var|ListGroup1]),
	findall(X,(member(V,ListGroup2),member(var(V,OwnVGroup),VarGroups),member(X,OwnVGroup)),ListGroup3),
	append(ListGroup2,ListGroup3,ListGroup4),
	remove_dups(ListGroup4,ListGroup).


event_type(EventOID,I,Negative_listIn,Negative_listOut,Event_type) :- 
	memberchk(var(I,[EventOID],Action),Negative_listIn), !,
	delete(Negative_listIn,var(I,[EventOID],Action),Negative_listOut),
	get_event_type(negative_event,Event_type).
event_type(_,_,Negative_listIn,Negative_listIn,Event_type) :- 
	get_event_type(positive_event,Event_type).

replace_negative_list(Event_type,I,Negative_listIn,Replaced_list,Replacement,Negative_listOut) :- 
	my_subclass_of(Event_type,mm_positive_event), !,
	member(var(I,EventList,Action),Negative_listIn),
	replace_last_element(EventList,Replaced_list,Replacement,NewEventList),
	replace_element(Negative_listIn,var(I,EventList,Action),var(I,NewEventList,Action),Negative_listOut).
replace_negative_list(_,_,Negative_listIn,_,_,Negative_listIn).


replace_event(_,[],_,_,[]) :- !.
replace_event(VarGroup,[var(Var,Event_list,Action)|RestIn],OldEvents,
	         NewEvent,[var(Var,NewEvent_list,Action)|RestOut]) :- 
	memberchk(Var,VarGroup), !,
	replace_last_event(Event_list,OldEvents,NewEvent,NewEvent_list),
	replace_event(VarGroup,RestIn,OldEvents,NewEvent,RestOut).
replace_event(VarGroup,[H|RestIn],OldEvents,NewEvent,[H|RestOut]) :- 
	replace_event(VarGroup,RestIn,OldEvents,NewEvent,RestOut).

/*replace_last_event(TempOldEvents,[],ev(NEV,NAEV,_,OutToken),NewEventList) :- !,
	remove_multiple_occs(TempOldEvents,ev(NEV,NAEV,_,OutToken),NewEventList).*/
replace_last_event(NewEventList,[],_,NewEventList) :- !.
replace_last_event(OldEvents,[ev(EV,AEV,_,InToken)|Tail_to_be_deleted],ev(NEV,NAEV,_,OutToken),NewEventList) :- 
	replace_element(OldEvents,ev(EV,AEV,ActualOP,InToken),ev(NEV,NAEV,ActualOP,OutToken),NextOldEvents),
	replace_last_event(NextOldEvents,Tail_to_be_deleted,ev(NEV,NAEV,_,OutToken),NewEventList).

replace_top_event(TempOldEvents,[],_,NewEventList) :- !,
	remove_dups(TempOldEvents,NewEventList).
replace_top_event(OldEvents,[OldEvent|Tail_to_be_deleted],NewEvent,NewEventList) :- 
	replace_element(OldEvents,OldEvent,NewEvent,NextOldEvents),
	replace_top_event(NextOldEvents,Tail_to_be_deleted,NewEvent,NewEventList).

remove_multiple_occs([],_,[]) :- !.
remove_multiple_occs([ev(NEV,NAEV,_,OutToken)|TailOldEvents],ev(NEV,NAEV,_,OutToken),NewEventList) :- 
	member(ev(NEV,NAEV,_,OutToken),TailOldEvents), !,
	remove_multiple_occs(TailOldEvents,ev(NEV,NAEV,_,OutToken),NewEventList).
remove_multiple_occs([H|TailOldEvents],NewEvent,[H|NewEventTail]) :- 
	remove_multiple_occs(TailOldEvents,NewEvent,NewEventTail).


replace_unique(_,[],_,_,[]) :- !.
replace_unique(VarGroup,[[Var,Event]|TailUniqueIn],OldEvents,CEventOID,[[Var,CEventOID]|TailUniqueOut]) :- 
	member(Var,VarGroup),
	member(Event,OldEvents), !,
	replace_unique(VarGroup,TailUniqueIn,OldEvents,CEventOID,TailUniqueOut).
replace_unique(VarGroup,[H|TailUniqueIn],OldEvents,CEventOID,[H|TailUniqueOut]) :- 
	replace_unique(VarGroup,TailUniqueIn,OldEvents,CEventOID,TailUniqueOut).


event_list_intersection([],_,[]) :- !.
event_list_intersection([ev(EV,AEV,OP,Token)|T1],L2,[ev(EV,AEV,OP,Token)|TCommon]) :- 
	member(ev(EV,AEV,_,Token),L2), !,
	event_list_intersection(T1,L2,TCommon).
event_list_intersection([_|T1],L2,LCommon) :- 
	event_list_intersection(T1,L2,LCommon).

second_pass_stage2([ev(Last_event_OID,Last_event_OID,_)],_,0,Last_event_OID) :- !.
second_pass_stage2([ev(Prim_event_OID,Prim_anti_event_OID,_)],_,0,[Prim_event_OID,Prim_anti_event_OID]) :- !.
second_pass_stage2([ev(EV1,AEV1,Left_in_token),ev(EV2,AEV2,Right_in_token)|RestTopEvents],
			Negative_listIn,No_of_events,Last_event_OID) :- 
	event_type(EV1,_,Negative_listIn,Negative_listNext1,Event_type1),	
	(Event_type1 == mm_negative_event ->
		(create_cartesian_event(Event_type1,EV2,AEV2,Right_in_token,
					  EV1,AEV1,Left_in_token,CEventOID,Out_token),
		 replace_negative_list(Event_type1,_,Negative_listNext1,[EV2,EV1],CEventOID,Negative_listNext));
		(event_type(EV2,_,Negative_listIn,Negative_listNext2,Event_type2), 
		 create_cartesian_event(Event_type2,EV1,AEV1,Left_in_token,
					  EV2,AEV2,Right_in_token,CEventOID,Out_token),
		 replace_negative_list(Event_type2,_,Negative_listNext2,[EV1,EV2],CEventOID,Negative_listNext))),
	second_pass_stage2([ev(CEventOID,CEventOID,Out_token)|RestTopEvents],
				Negative_listNext,PreviousNo_of_events,Last_event_OID),
	No_of_events is PreviousNo_of_events + 1.


/* The rest are predicates that create actual objects in EXACT */

create_primitive_event(Prefix,Attr,Class,Vars,PEventOID,Token) :- 
	get_event_type(inheritable_ms_event,Event_type),
	atom_chars(Attr,AttrStr),
	append(Prefix,AttrStr,MethodStr), 
	atom_chars(Method,MethodStr),
	transform_vars(Vars,Token), 
	% Check if event already exists
	((get_by_active_method([Method],PEventOID) => mm_inheritable_ms_event,
	  get_when(after) => PEventOID,
	  get_active_class(Class) => PEventOID) ->           
		 (get_out_token(token(_O1,_O2,Token)) => PEventOID -> 
			true;
			put_out_token([token([],[],Token)]) => PEventOID);
		 new([PEventOID,[active_method([Method]), active_class([Class]), when([after]),
		                 out_token([token([],[],Token)])
	         ]]) => Event_type),
	create_primitive_call(Prefix,Attr,Class,Token,PEventOID).

create_primitive_call("put_",Attribute,Class,Token,Event) :- !,
	concat_atoms(get_,Attribute,Get_name),
	((Token = [var(_,[OID]),var(_,[Val])],
	  Method =.. [Get_name,Val],
	  Call = (get(OID) => Class, Method => OID));
	 (Token = [var(_,[OID]),var(_,[Val1]),var(_,[Val2])],
	  Method =.. [Get_name,Val1],
	  Call = (get(OID) => Class, Method => OID, Val2=Val1))),
	(get_primitive_call([TokenTail,Call]) => Event -> 
		(member(Token,TokenTail) -> 
			true;
			update_primitive_call([[TokenTail,Call],[[Token|TokenTail],Call]]) => Event);
		put_primitive_call([[[Token],Call]]) => Event).
create_primitive_call(_,_,_,_,_).


create_update_event(Attr,Class,Out_token,EventOID,AntiEventOID) :- 
	get_event_type(inheritable_ms_event,Event_type),
	atom_chars(Attr,AttrStr),
	append("update_",AttrStr,MethodStr),   % PIPE
	atom_chars(Method,MethodStr),
	% Check if event already exists
	((get_by_active_method([Method],PEventOID) => mm_inheritable_ms_event,
	  get_when(after) => PEventOID,
	  get_active_class(Class1) => PEventOID,
	  (my_subclass_of(Class,Class1) ->  % If the new class is a subclass of an existing event do nothing 
		true;
		(my_subclass_of(Class1,Class) -> % If the new class is a superclass of an existing event, then upgrade the class
			update_active_class([Class1,Class]) => PEventOID;
			fail))) ->           
	         (get_out_token(token(_,_,Out_token)) => PEventOID -> 
		         true;
			 (get_out_token(token([[EventOID,AntiEventOID]],[],Previous_Out_token)) => PEventOID  -> 
				(union(Previous_Out_token,Out_token,New_out_token),
				 replace_out_token([token([[EventOID,AntiEventOID]],[],Previous_Out_token),
							token([[EventOID,AntiEventOID]],[],New_out_token)]) => PEventOID);
			 	put_out_token([token([[EventOID,AntiEventOID]],[],Out_token)]) => PEventOID));
		 new([PEventOID,[active_method([Method]), active_class([Class]), when([after]),
		                 out_token([token([[EventOID,AntiEventOID]],[],Out_token)])
	         ]]) => Event_type).

create_class_event(Method,Class,Vars,PEventOID,Token) :- 
	get_event_type(inheritable_ms_event,Event_type),
	transform_vars(Vars,Token),
	((get_by_active_method([Method],PEventOID) => mm_inheritable_ms_event,
	  get_active_class(Class) => PEventOID,
	  get_when(after) => PEventOID) ->            % Check if event already exists
		 (get_out_token(token(_O1,_O2,Token)) => PEventOID -> 
			true;
			put_out_token([token([],[],Token)]) => PEventOID);
		 new([PEventOID,[active_method([Method]), active_class([Class]), when([after]),
		                 out_token([token([],[],Token)])
	         ]]) => Event_type).



create_logical_event(Needed,OID_Var,Var,In_token,PEventOID,AntiPEventOID,Preds,LEventOID,Out_token) :- 
	get_event_type(logical_event,Event_type),
	get_out_token(token(L1,L2,In_token)) => PEventOID, 
	get_out_token(token(AL1,AL2,In_token)) => AntiPEventOID, 	
	(Needed == 1 -> 
		Out_token = In_token;
		delete(In_token,var(Var,[Val]),Out_token)),
	create_logical_predicate(Preds,Val,Pred),
	((get_by_predicate([[_,Pred]],LEventOID1) => mm_logical_event,         % Check if a logical event already exists
	  ((get_in_token(token(EVs,AEVs,In_token)) => LEventOID1, 
	    member(PEventOID,EVs), member(AntiPEventOID,AEVs)) -> 
		         (NewInToken = In_token,
			  LEventOID = LEventOID1);
			 ((get_in_token(token(EVs,AEVs,OtherIn_token)) => LEventOID1,
			   OtherIn_token = [var(OID_Var,[_])|Rest1],
			   In_token = [var(OID_Var,[_])|Rest2],
		   	   member(PEventOID,EVs), member(AntiPEventOID,AEVs)) -> 
				(append(Rest1,Rest2,Rest),
				 NewInToken = [var(OID_Var,[_])|Rest],
				 update_in_token([token(EVs,AEVs,OtherIn_token),
						  token(EVs,AEVs,NewInToken)]) => LEventOID1,
				 delete_out_token([token([LEventOID],[],OtherIn_token)]) => PEventOID,
				 delete_out_token([token([],[LEventOID],OtherIn_token)]) => AntiPEventOID,
				 replace_out_token([token(L1,L2,In_token),token(L1,L2,NewInToken)]) => PEventOID,
				 replace_out_token([token(AL1,AL2,In_token),token(AL1,AL2,NewInToken)]) => AntiPEventOID,
				 LEventOID = LEventOID1);
		         	((get_in_token(token(EVs,AEVs,_)) => LEventOID1, 
		   		  member(PEventOID,EVs), member(AntiPEventOID,AEVs)) -> 
					(put_in_token([token([PEventOID],[AntiPEventOID],In_token)]) => LEventOID1,
					 LEventOID = LEventOID1,
					 NewInToken = In_token);
					fail)
				  )));
	 true),
	(var(LEventOID) -> 
		(new([LEventOID,[in_token([token([PEventOID],[AntiPEventOID],In_token)]), 
		                out_token([token([],[],Out_token)]), predicate([[Val,Pred]])
	         ]]) => Event_type,
		 NewInToken = In_token);
		(get_out_token(token(_,_,Out_token)) => LEventOID ->
		         true;
			 put_out_token([token([],[],Out_token)]) => LEventOID)),
	(member(LEventOID,L1) ->
	        true;
		(replace_out_token([token(L1,L2,NewInToken),token([LEventOID|L1],L2,NewInToken)]) => PEventOID,
		 replace_out_token([token(AL1,AL2,NewInToken),token(AL1,[LEventOID|AL2],NewInToken)]) => AntiPEventOID)).


create_join_event(Event_type,Var,EV1,AEV1,OP1,Left_in_token,EV2,AEV2,OP2,Right_in_token,
			CEventOID,Next_op,Out_token) :- 
	get_out_token(token(LL1,LL2,Left_in_token)) => EV1,
	(AEV1 \== EV1 -> 
	    get_out_token(token(ALL1,ALL2,Left_in_token)) => AEV1;
	    (ALL1 = LL1, ALL2 = LL2)),
	get_out_token(token(RL1,RL2,Right_in_token)) => EV2,
	(AEV2 \== EV2 -> 
	    get_out_token(token(ARL1,ARL2,Right_in_token)) => AEV2;
	    (ARL1 = RL1, ARL2 = RL2,
	     (Event_type == mm_negative_event ->
	     	(get_deletion_optimization(RightOpt) => EV2,
	     	 (RightOpt == on -> 
	     	 	update_deletion_optimization([on,off]) => EV2;
	     	 	true));
	     	 true))),
	create_comparison(OP1,OP2,LR_OP,OUT_CARD),
	create_comparison(OP2,OP1,RL_OP,_), !,
	create_value(OP1,LVal),
	create_value(OP2,RVal),
	deletion_optimization(OP1,OP2,OPT),
	(OUT_CARD < 3 -> 
		Next_op = ['='];
		append(OP1,OP2,Next_op)),
	create_join_predicates(LR_OP,RL_OP,LVal,RVal,LR_pred,RL_pred),
	create_out_value(Event_type,OUT_CARD,LVal,RVal,OutVal),
          % Check if the two-input event already exists
	find_existing_join_event(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),
				EV1,AEV1,Left_in_token,Left_old,LeftSide,
				EV2,AEV2,Right_in_token,RightSide,Right_old,
				Event_type,CEventOID), !,
	((var(LeftSide), var(RightSide), % If the two inputs are the same then re-use the join event by extending the join!
	  EV1 == EV2, AEV1 == AEV2) -> 
		(CEventOID = EV1,
		 get_join(Join_var_list) => CEventOID,
		 (\+ member(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),Join_var_list) ->
		 	update_join([Join_var_list,[var(Var,OutVal,LVal,RVal,LR_pred,RL_pred)|Join_var_list]]) => CEventOID;
			true),
		 (get_deletion_optimization(on) => CEventOID, OPT == off -> 
			update_deletion_optimization([on,off]) => CEventOID;
			true),
		 get_out_token(token(_,_,OUT1)) => CEventOID,
	         (OUT_CARD < 3 -> 
			Out_token = OUT1;
			replace_element(OUT1,var(Var,OldTemplate),var(Var,[_|OldTemplate]),Out_token)));
	((var(LeftSide), var(RightSide),
	  member(CEventOID,LL1), member(CEventOID,RL1),   % If the two input events have been re-joined then re-use
	  member(CEventOID,ALL2), member(CEventOID,ARL2)) -> % the previous two_input event by extending the join!
		(get_join(Join_var_list) => CEventOID,
		 (\+ member(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),Join_var_list) ->
		 	update_join([Join_var_list,[var(Var,OutVal,LVal,RVal,LR_pred,RL_pred)|Join_var_list]]) => CEventOID;
			true),
		 (get_deletion_optimization(on) => CEventOID, OPT == off -> 
			update_deletion_optimization([on,off]) => CEventOID;
			true),
		 get_out_token(token(_,_,OUT1)) => CEventOID,
	         (OUT_CARD < 3 -> 
			Out_token = OUT1;
			replace_element(OUT1,var(Var,OldTemplate),var(Var,[_|OldTemplate]),Out_token)));
	% Else everything else! No need for updating tokens anywhere!
	((nonvar(LeftSide), nonvar(RightSide)) ->       % The 2-input event already exists and can be re-used
	    ((same_set(Left_in_token,Left_old) -> 
		   true;
		   ((get_left_in_token(token([EV1],[AEV1],Left_in_token)) => CEventOID ->
			 true;
		 	 (put_left_in_token([token([EV1],[AEV1],Left_in_token)]) => CEventOID,
			  ((EV1 == AEV1, get_left_virtual(virtual(true,GL)) => CEventOID) ->
				update_left_virtual([virtual(true,GL),virtual(false,true)]) => CEventOID;
				true))),
		    (EV1 \== AEV1 ->
		         (replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],LL2,Left_in_token)]) => EV1,
		          replace_out_token([token(ALL1,ALL2,Left_in_token),token(ALL1,[CEventOID|ALL2],Left_in_token)]) => AEV1);
			 replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],[CEventOID|LL2],Left_in_token)]) => EV1))),
	      (same_set(Right_in_token,Right_old) -> 
		    true;
		    ((get_right_in_token(token([EV2],[AEV2],Right_in_token)) => CEventOID ->
			  true;
			  (put_right_in_token([token([EV2],[AEV2],Right_in_token)]) => CEventOID,
			   ((EV2 == AEV2, get_right_virtual(virtual(true,GR)) => CEventOID) ->
				update_right_virtual([virtual(true,GR),virtual(false,true)]) => CEventOID;
				true))),
		     (EV2 \== AEV2 ->
		         (replace_out_token([token(RL1,RL2,Right_in_token),token([CEventOID|RL1],RL2,Right_in_token)]) => EV2,
		          replace_out_token([token(ARL1,ARL2,Right_in_token),token(ARL1,[CEventOID|ARL2],Right_in_token)]) => AEV2);
			  replace_out_token([token(RL1,RL2,Right_in_token),
					     token([CEventOID|RL1],[CEventOID|RL2],Right_in_token)]) => EV2))),
	     (my_subclass_of(Event_type,mm_positive_event) ->
	                    union(Left_in_token,Right_in_token,OUT1);
			    OUT1 = Left_in_token),
	     (OUT_CARD < 3 -> 
		Out_token = OUT1;
		replace_element(OUT1,var(Var,OldTemplate),var(Var,[_|OldTemplate]),Out_token)),
	     (get_out_token(token(O1,O2,OUT2)) => CEventOID -> 
	       	    (same_set(Out_token,OUT2) -> 
	            	    true;
		            ((O1 == [], O2 == []) -> 
			         replace_out_token([token([],[],OUT2),token([],[],Out_token)]) => CEventOID;
				 put_out_token([token([],[],Out_token)]) => CEventOID));
		    put_out_token([token([],[],Out_token)]) => CEventOID));

	    % An existing 2-input event can be re-used by extending its join slot!
	(re_use(EV1,AEV1,EV2,AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN) ->              
		((get_join(Join_var_list) => CEventOID -> 
			((\+ member(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),Join_var_list) ->
		 		update_join([Join_var_list,[var(Var,OutVal,LVal,RVal,LR_pred,RL_pred)|Join_var_list]]) => CEventOID;
				true),
		 	 (get_deletion_optimization(on) => CEventOID, OPT == off -> 
				update_deletion_optimization([on,off]) => CEventOID;
				true));
			put_join([[var(Var,OutVal,LVal,RVal,LR_pred,RL_pred)]]) => CEventOID),			  
		 (OUT_CARD < 3 -> 
		 	Out_token = OUTTOKEN;
			replace_element(OUTTOKEN,var(Var,OldTemplate),var(Var,[_|OldTemplate]),Out_token)));
				      % The 2-input does not exist, so create it!
	                              ((my_subclass_of(Event_type,mm_positive_event) ->
					           union(Left_in_token,Right_in_token,OUT);
						   OUT = Left_in_token),
				       (OUT_CARD < 3 -> 
		 				Out_token = OUT;
						replace_element(OUT,var(Var,OldTemplate),
							        var(Var,[_|OldTemplate]),Out_token)),
				       create_in_token(EV1,Left_in_token,Var,Left_in_tokenExt),
				       create_in_token(EV2,Right_in_token,Var,Right_in_tokenExt),
				       create_virtual_node(EV1,AEV1,Left_in_token,VirtualNodeL),
				       create_virtual_node(EV2,AEV2,Right_in_token,VirtualNodeR),
				       create_counter_info(EV1,LeftCounter),
				       create_counter_info(EV2,RightCounter),		       
				       new([CEventOID,[
						left_in_token([token([EV1],[AEV1],Left_in_tokenExt)]), 
						left_virtual([VirtualNodeL]),
						left_counter([LeftCounter]),
				                right_in_token([token([EV2],[AEV2],Right_in_tokenExt)]),
						right_virtual([VirtualNodeR]),
						right_counter([RightCounter]),
		                                join([[var(Var,OutVal,LVal,RVal,LR_pred,RL_pred)]]),
						deletion_optimization([OPT]),
						out_token([token([],[],Out_token)])
	                               ]]) => Event_type,
	                               (EV1 == EV2 ->
	                               		replace_out_token([token(LL1,LL2,Left_in_token),
							token([CEventOID|LL1],[CEventOID|LL2],Left_in_token)]) => EV1;
				       ((EV1 \== AEV1 ->
					   (replace_out_token([token(LL1,LL2,Left_in_token),
								token([CEventOID|LL1],LL2,Left_in_token)]) => EV1,
					    replace_out_token([token(ALL1,ALL2,Left_in_token),
								token(ALL1,[CEventOID|ALL2],Left_in_token)]) => AEV1);
					   replace_out_token([token(LL1,LL2,Left_in_token),
								token([CEventOID|LL1],[CEventOID|LL2],Left_in_token)]) => EV1),
				       (EV2 \== AEV2 ->
					   (replace_out_token([token(RL1,RL2,Right_in_token),
								token([CEventOID|RL1],RL2,Right_in_token)]) => EV2,
					    replace_out_token([token(ARL1,ARL2,Right_in_token),
								token(ARL1,[CEventOID|ARL2],Right_in_token)]) => AEV2);
					    replace_out_token([token(RL1,RL2,Right_in_token),
							       token([CEventOID|RL1],[CEventOID|RL2],Right_in_token)]) => EV2))))))
	)).
create_join_event(Event_type,_,EV1,AEV1,OP1,Left_in_token,EV2,AEV2,OP2,Right_in_token,CEventOID,Next_op,Out_token) :- 
	append(OP1,OP2,Next_op),
	create_cartesian_event(Event_type,EV1,AEV1,Left_in_token,EV2,AEV2,Right_in_token,CEventOID,Out_token).


find_existing_join_event(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),
				EV1,AEV1,Left_in_token,Left_old,LeftSide,
				EV2,AEV2,Right_in_token,RightSide,Right_old,
				Event_type,CEventOID) :-
	((get(CEventOID) => Event_type,
	  get_join(Join_var_list) => CEventOID,
	  member(var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),Join_var_list),
          ((get_left_in_token(token(Left,ALeft,Left_oldTempl)) => CEventOID,
	    templ_to_token(Left_oldTempl,Left_old),
	    member(EV1,Left), member(AEV1,ALeft),
	    same_set(Left_old,Left_in_token)) -> 
	              LeftSide = true),         % Both inputs must be the same with the same token as well.
          ((get_right_in_token(token(Right,ARight,Right_oldTempl)) => CEventOID,
	    templ_to_token(Right_oldTempl,Right_old),
	    member(EV2,Right), member(AEV2,ARight),
	    same_set(Right_old,Right_in_token)) -> 
	              RightSide = true));
	 true).

create_cartesian_event(Event_type,EV1,AEV1,Left_in_token,EV2,AEV2,Right_in_token,CEventOID,Out_token) :- 
	get_out_token(token(LL1,LL2,Left_in_token)) => EV1, 
	(AEV1 \== EV1 -> 
	    get_out_token(token(ALL1,ALL2,Left_in_token)) => AEV1;
	    true),
	get_out_token(token(RL1,RL2,Right_in_token)) => EV2, 
	(AEV2 \== EV2 -> 
	    get_out_token(token(ARL1,ARL2,Right_in_token)) => AEV2;
	    true),
	find_existing_cartesian_event(EV1,AEV1,Left_in_token,Left_old,LeftSide,
					EV2,AEV2,Right_in_token,RightSide,Right_old,
					Event_type,CEventOID), !,
	((nonvar(LeftSide), nonvar(RightSide)) ->       % The 2-input event already exists and can be re-used
	    ((same_set(Left_in_token,Left_old) -> 
		   true;
		   ((get_left_in_token(token([EV1],[AEV1],Left_in_token)) => CEventOID ->
			 true;
		 	 (put_left_in_token([token([EV1],[AEV1],Left_in_token)]) => CEventOID,
			  ((EV1 == AEV1, get_left_virtual(virtual(true,GL)) => CEventOID) ->
				update_left_virtual([virtual(true,GL),virtual(false,true)]) => CEventOID;
				true))),
		    (EV1 \== AEV1 ->
		         (replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],LL2,Left_in_token)]) => EV1,
		          replace_out_token([token(ALL1,ALL2,Left_in_token),token(ALL1,[CEventOID|ALL2],Left_in_token)]) => AEV1);
			 replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],[CEventOID|LL2],Left_in_token)]) => EV1))),
	      (same_set(Right_in_token,Right_old) -> 
		    true;
		    ((get_right_in_token(token([EV2],[AEV2],Right_in_token)) => CEventOID ->
			  true;
			  (put_right_in_token([token([EV2],[AEV2],Right_in_token)]) => CEventOID,
			   ((EV2 == AEV2, get_right_virtual(virtual(true,GR)) => CEventOID) ->
				update_right_virtual([virtual(true,GR),virtual(false,true)]) => CEventOID;
				true))),
		     (EV2 \== AEV2 ->
		         (replace_out_token([token(RL1,RL2,Right_in_token),token([CEventOID|RL1],RL2,Right_in_token)]) => EV2,
		          replace_out_token([token(ARL1,ARL2,Right_in_token),token(ARL1,[CEventOID|ARL2],Right_in_token)]) => AEV2);
			 replace_out_token([token(RL1,RL2,Right_in_token),token([CEventOID|RL1],[CEventOID|RL2],Right_in_token)]) => EV2))),
	     union(Left_in_token,Right_in_token,Out_token),
	     (get_out_token(token(O1,O2,OUT2)) => CEventOID -> 
	       	    (same_set(Out_token,OUT2) -> 
	            	    true;
		            ((O1 == [], O2 == []) -> 
				 replace_out_token([token([],[],OUT2),token([],[],Out_token)]) => CEventOID;
				 true));
		    put_out_token([token([],[],Out_token)]) => CEventOID));
	    % The 2-input does not exist, so create it!
	    ((my_subclass_of(Event_type,mm_positive_event) ->
			union(Left_in_token,Right_in_token,Out_token);
			Out_token = Left_in_token),
	     create_in_token(EV1,Left_in_token,[],Left_in_tokenExt),
	     create_in_token(EV2,Right_in_token,[],Right_in_tokenExt),
	     create_virtual_node(EV1,AEV1,Left_in_token,VirtualNodeL),
	     create_virtual_node(EV2,AEV2,Right_in_token,VirtualNodeR),
	     create_counter_info(EV1,LeftCounter),
	     create_counter_info(EV2,RightCounter),
	     new([CEventOID,[left_in_token([token([EV1],[AEV1],Left_in_tokenExt)]), left_virtual([VirtualNodeL]),
			     left_counter([LeftCounter]), right_counter([RightCounter]),
	                     right_in_token([token([EV2],[AEV2],Right_in_tokenExt)]), right_virtual([VirtualNodeR]),
			     deletion_optimization([on]), out_token([token([],[],Out_token)])
	     ]]) => Event_type,
	     (EV1 \== AEV1 ->
		(replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],LL2,Left_in_token)]) => EV1,
		 replace_out_token([token(ALL1,ALL2,Left_in_token),token(ALL1,[CEventOID|ALL2],Left_in_token)]) => AEV1);
		replace_out_token([token(LL1,LL2,Left_in_token),token([CEventOID|LL1],[CEventOID|LL2],Left_in_token)]) => EV1),
	     (EV2 \== AEV2 ->
		(replace_out_token([token(RL1,RL2,Right_in_token),token([CEventOID|RL1],RL2,Right_in_token)]) => EV2,
		 replace_out_token([token(ARL1,ARL2,Right_in_token),token(ARL1,[CEventOID|ARL2],Right_in_token)]) => AEV2);
		replace_out_token([token(RL1,RL2,Right_in_token),token([CEventOID|RL1],[CEventOID|RL2],Right_in_token)]) => EV2))).


find_existing_cartesian_event(	EV1,AEV1,Left_in_token,Left_old,LeftSide,
				EV2,AEV2,Right_in_token,RightSide,Right_old,
				Event_type,CEventOID) :-

	((get(CEventOID) => Event_type,           % Check if the two-input event already exists
	  \+ (get_join(_) => CEventOID),  
          ((get_left_in_token(token(Left,ALeft,Left_oldTempl)) => CEventOID,
	    templ_to_token(Left_oldTempl,Left_old),
	    member(EV1,Left), member(AEV1,ALeft),
	    same_set(Left_old,Left_in_token)) -> 
	              LeftSide = true),         % Both inputs must be the same with the same token as well.
          ((get_right_in_token(token(Right,ARight,Right_oldTempl)) => CEventOID,
	    templ_to_token(Right_oldTempl,Right_old),
	    member(EV2,Right), member(AEV2,ARight),
	    %either_way_subset(Right_old,Right_in_token)) -> 
	    same_set(Right_old,Right_in_token)) -> 
	              RightSide = true));
	 true).



create_rule(Last_event_OID,PrologGoal,Action,AuxAtts,Total_events,CondVars,RuleOID) :- 
	(is_list(Last_event_OID) -> 
		Last_event_OID = [Event,AntiEvent];
		(Event = Last_event_OID,
		 AntiEvent = Last_event_OID)),
	get_out_token(token(_L1,_L2,Token)) => Event,
	my_subset(CondVars,Token),
	(Event \== AntiEvent -> 
		get_out_token(token(_LL1,_LL2,Token)) => AntiEvent;
		true),
	append(Action,[46],Action1),	% add a dot at the end: FOTIS
	open_chars_stream(Action1,Stream1),
	read_term(Stream1,ActionTerm,[variable_names(VarList1)]),

	close(Stream1),
	unify_vars_in_token(VarList1,Token),   % Unifies Action-Condition

	(PrologGoal \== true -> 

		(

		append(PrologGoal,[46],PrologGoal1),
		open_chars_stream(PrologGoal1,Stream2),
		read_term(Stream2,PrologGoalTerm,[variable_names(VarList2)]),

		 close(Stream2),
		 var_token(VarList2,CRToken),
		 unify_vars_in_token(VarList1,CRToken), % Unifies Action-Prolog Goal
		 union(Token,CRToken,_),                % Unifies Condition-Prolog Goal
                 ConditionGoal = (occurrence_params(CondVars),condition_result(CRToken),PrologGoalTerm),
		 ActionGoal = (occurrence_params(CondVars),condition_result(CRToken),ActionTerm));      
		(ConditionGoal = true,
		 ActionGoal = (occurrence_params(CondVars),ActionTerm))),
	append([event([Event]),anti_event([AntiEvent]),
		condition([ConditionGoal]),action([ActionGoal]),
		in_token([CondVars]),event_number([Total_events])],AuxAtts,Attributes),
	new([RuleOID,Attributes]) => super.




% This clause catters for the case where a new or delete event (without attributes) are traced!
create_virtual_node(Event,AntiEvent,SToken,virtual(true,[Goal1,Goal3])) :- 
	Event \== AntiEvent,
	my_instance_of(Event,mm_inheritable_ms_event),
	my_instance_of(AntiEvent,mm_inheritable_ms_event),
	get_active_method(new) => Event, !,
	get_active_class(Class) => Event,
	copy_term(SToken,SToken1),
	SToken1 = [var(OIDVar,[OID])],
	% The join is on the OID!
	Goal1 = [(occurrence_params(In_token,SToken1,_,_),
		  member(var(OIDVar,[OID]),In_token),
		  nonvar(OID)),   % If the OID is in common vars then bind it!
		 true],            
	Goal3 = [occurrence_params(_,SToken1,_,_),
		 (get(OID) => Class)].
% This is the usual! case
create_virtual_node(Event,AntiEvent,SToken,virtual(true,[Goal1,Goal2,Goal3])) :- 
	Event \== AntiEvent,
	my_instance_of(Event,mm_inheritable_ms_event),
	my_instance_of(AntiEvent,mm_inheritable_ms_event), !,
	get_active_method(Put_name) => Event,
	atom_chars(Put_name,SPut_name),
	append("put_",Attribute,SPut_name),   % PIPE
	append("get_",Attribute,SGet_name),
	atom_chars(Get_name,SGet_name),
	append("get_by_",Attribute,SGet_By_name),
	atom_chars(Get_By_name,SGet_By_name),
	get_active_class(Class) => Event,
	MethodGet =.. [Get_name,Val],
	MethodGetBy =.. [Get_By_name,[Val],OID],
	copy_term(SToken,SToken1),
	((SToken1 = [var(OIDVar,[OID]),var(_Var,[Val])], !);
	 SToken1 = [var(OIDVar,[OID]),var(_Var1,[Val]),var(_Var2,[Val])]),
	% The join is on the OID!
	Goal1 = [(occurrence_params(In_token,SToken1,_,_),
		  member(var(OIDVar,[OID]),In_token),
		  nonvar(OID)),   % If the OID is in common vars then bind it!
		 MethodGet => OID],                    % Get the rest vars
	Goal2 = [(occurrence_params(In_token,SToken1,JoinVarList,Predicates),
		  nth0(Index,JoinVarList,var(Var,_,_,_,_,_)),
		  nth0(Index,Predicates,_=_),       % Is the join on the attribute an equi-join?
		  member(var(Var,[Val]),In_token),
		  nonvar(Val)), % Bind it!
		 MethodGetBy => Class],		  % Get the OIDs of the objects with that attribute!
	Goal3 = [occurrence_params(_,SToken1,_,_),
		 (get(OID) => Class, 
		  MethodGet => OID)].
% The following creates virtual two-input nodes. However it does not speed-up things!
create_virtual_node(Event,Event,SToken,virtual(true,AllGoals)) :- 
	my_instance_of(Event,mm_two_input_event), 
	get_right_virtual(virtual(true,RightGoals)) => Event,
	get_left_virtual(virtual(true,LeftGoals)) => Event,	
	append([[(occurrence_params(In_token,STokenR,_,_),Guard1),CallR1]|MiddleGoalsR],
		[[occurrence_params(_,STokenR,_,_),(ClassCall,CallR1)]],
		RightGoals),
	append([[(occurrence_params(In_token,STokenL,_,_),Guard1),CallL1]|MiddleGoalsL],
		[[occurrence_params(_,STokenL,_,_),(ClassCall,CallL1)]],
		LeftGoals),
	!, % Both inputs are virtual nodes and the Event is joined on the OIDVar!
	   % The same Guard1 call assures that the same variable is used to join both inputs
	copy_term(SToken,SToken1),
	my_subset(STokenR,SToken1),
	my_subset(STokenL,SToken1),
	GoalFirst = [(occurrence_params(In_token,SToken1,_,_),Guard1),(CallL1,CallR1)],
	construct_middle_goals(In_token,SToken1,STokenL,MiddleGoalsL,CallR1,MiddleGoalsLR),
	construct_middle_goals(In_token,SToken1,STokenR,MiddleGoalsR,CallL1,MiddleGoalsRL),
	GoalLast = [occurrence_params(_,SToken1,_,_),(ClassCall,CallL1,CallR1)],
	append([GoalFirst|MiddleGoalsLR],MiddleGoalsRL,Goals1),
	append(Goals1,[GoalLast],AllGoals).
create_virtual_node(_,_,_,virtual(false,true)).


construct_middle_goals(_,_,_,[],_,[]) :- !.
construct_middle_goals(In_token,SToken1,STokenLorR,
	[[(occurrence_params(In_token,STokenLorR,JoinVarList,Predicates),GuardL2orR2),
	   CallL2orR2]|MiddleGoalsLorR_tail],
	CallR1orL1,
	[[(occurrence_params(In_token,SToken1,JoinVarList,Predicates),GuardL2orR2),
	   (CallL2orR2,CallR1orL1)]|MiddleGoalsLRorRL_tail]) :- 
   construct_middle_goals(In_token,SToken1,STokenLorR,MiddleGoalsLorR_tail,
			  CallR1orL1,MiddleGoalsLRorRL_tail).

%create_in_token(EV1,Left_in_token,Var,Left_in_tokenExt) :- 
create_in_token(Event,In_token,JoinVar,[var(JoinVar,[Val],true,true)|In_tokenExtTail]) :- 
	(my_instance_of(Event,mm_inheritable_ms_event);
	 my_instance_of(Event,mm_logical_event)),
	In_token = [var(JoinVar,[Val])|Rest], !,
	create_non_stored_vars(Event,Rest,JoinVar,Val,In_tokenExtTail).
create_in_token(Event,In_token,JoinVar,[var(OIDVar,[Val],true,true)|In_tokenExtTail]) :- 
	(my_instance_of(Event,mm_inheritable_ms_event);
	 my_instance_of(Event,mm_logical_event)), !,
	In_token = [var(OIDVar,[Val])|Rest],
	create_non_stored_vars(Event,Rest,OIDVar,Val,In_tokenExtTail1),
	replace_element(In_tokenExtTail1,var(JoinVar,Val1,false,Code),
			var(JoinVar,Val1,true,Code),In_tokenExtTail).
create_in_token(Event,In_token,JoinVar,In_tokenExt) :- 
	create_in_token_list(Event,In_token,JoinVar,In_tokenExt).

create_in_token_list(_,[],_,[]) :- !.
create_in_token_list(Event,[var(JoinVar,Val)|In_tokenTail],JoinVar,
			   [var(JoinVar,Val,true,Code)|In_tokenExtTail]) :- 
	(get_left_in_token(token(_,_,INToken)) => Event;
	 get_right_in_token(token(_,_,INToken)) => Event),
	(member(var(JoinVar,Val,false,Code),INToken);
	 member(var(JoinVar,_Val,false,Code),INToken)), !,
	create_in_token_list(Event,In_tokenTail,JoinVar,In_tokenExtTail).
create_in_token_list(Event,[var(Var,Val)|In_tokenTail],JoinVar,
			   [var(Var,Val,false,Code)|In_tokenExtTail]) :- 
	(get_left_in_token(token(_,_,INToken)) => Event;
	 get_right_in_token(token(_,_,INToken)) => Event),
	(member(var(Var,Val,false,Code),INToken);
	 member(var(Var,_Val,false,Code),INToken)), !,
	create_in_token_list(Event,In_tokenTail,JoinVar,In_tokenExtTail).
create_in_token_list(Event,[var(Var,Val)|In_tokenTail],JoinVar,[var(Var,Val,true,true)|In_tokenExtTail]) :- 
	(get_left_in_token(token(_,_,INToken)) => Event;
	 get_right_in_token(token(_,_,INToken)) => Event),
	(member(var(Var,Val,true,true),INToken);
	 member(var(Var,_Val,true,true),INToken)), !,
	create_in_token_list(Event,In_tokenTail,JoinVar,In_tokenExtTail).
create_in_token_list(Event,[var(JoinVar,Val)|In_tokenTail],JoinVar,
			   [var(JoinVar,Val,true,Code)|In_tokenExtTail]) :- 
	(get_left_in_token(token(_,_,INToken)) => Event;
	 get_right_in_token(token(_,_,INToken)) => Event),
	(member(var(JoinVar,Val,true,Code),INToken);
	 member(var(JoinVar,_Val,true,Code),INToken)), !,
	create_in_token_list(Event,In_tokenTail,JoinVar,In_tokenExtTail).
create_in_token_list(Event,[var(Var,Val)|In_tokenTail],JoinVar,
			   [var(Var,Val,false,Code)|In_tokenExtTail]) :- 
	(get_left_in_token(token(_,_,INToken)) => Event;
	 get_right_in_token(token(_,_,INToken)) => Event),
	(member(var(Var,Val,true,Code),INToken);
	 member(var(Var,_Val,true,Code),INToken)), !,
	create_in_token_list(Event,In_tokenTail,JoinVar,In_tokenExtTail).

%create_non_stored_vars(Event,Rest,Val,In_tokenExtTail)
create_non_stored_vars(_,[],_,_,[]) :- !.
create_non_stored_vars(Event,[var(Var,[Value])|TailVars],OIDVar,OIDVal,[var(Var,[Value],false,Code)|TailVarsExt]) :- 
	discover_attribute(Event,Attribute),
	concat_atoms(get_,Attribute,Method),
	MethodCall =.. [Method,Value],
	Code = (var(OIDVar,[OIDVal]), MethodCall => OIDVal),
	create_non_stored_vars(Event,TailVars,OIDVar,OIDVal,TailVarsExt).

discover_attribute(Event,Attribute) :- 
	my_instance_of(Event,mm_inheritable_ms_event), !,
	get_active_method(Method) => Event,
	reverse_method_name(Method,Attribute).
discover_attribute(Event,Attribute) :- 
	my_instance_of(Event,mm_logical_event),
	get_activated_by(PrimEvent) => Event,
	discover_attribute(PrimEvent,Attribute).

reverse_method_name(Method,Attribute) :- 
	atom_chars(Method,MethodS),
	(append("put_",AttributeS,MethodS);    % PIPE
	 append("delete_",AttributeS,MethodS);  % PIPE
	 append("update_",AttributeS,MethodS)), !,  % PIPE
	atom_chars(Attribute,AttributeS).


%create_counter_info(EV1,LeftCounter),
create_counter_info(Event,Counter) :- 
	my_instance_of(Event,mm_logical_event),
	% Only logical events which only check and do not deliver variables
	% can be considered with counter!
	get_out_token(token(_,_,[_])) => Event, !, 
	get_activated_by(PrimEvent) => Event,
	get_active_method(Method) => PrimEvent,
	get_active_class(Class) => PrimEvent,
	reverse_method_name(Method,Attribute),
	find_cardinality(Attribute,Class,S),
	(S == set ->
		Counter = on;
		Counter = off).
create_counter_info(Event,on) :- 
	my_instance_of(Event,mm_two_input_event),
	(get_left_counter(on) => Event;
	 get_right_counter(on) => Event), !.
create_counter_info(_Event,off).



re_use(EV1,AEV1,EV2,_AEV2,Left_in_token,_Right_in_token,EV2,OUTTOKEN) :- 
	get_left_in_token(token(Left,ALeft,Left_in_token1Templ)) => EV2,
	templ_to_token(Left_in_token1Templ,Left_in_token1),
	same_set(Left_in_token1,Left_in_token),
	member(EV1,Left), member(AEV1,ALeft), !,
	get_out_token(token(_,_,OUTTOKEN)) => EV2,
	my_subset(Left_in_token,OUTTOKEN).
re_use(EV1,_AEV1,EV2,AEV2,_Left_in_token,Right_in_token,EV1,OUTTOKEN) :- 
	get_right_in_token(token(Right,ARight,Right_in_token1Templ)) => EV1,
	templ_to_token(Right_in_token1Templ,Right_in_token1),
	same_set(Right_in_token1,Right_in_token),
	member(EV2,Right), member(AEV2,ARight), !,
	get_out_token(token(_,_,OUTTOKEN)) => EV1,
	my_subset(Right_in_token,OUTTOKEN).
re_use(EV1,AEV1,EV2,_AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN) :- 
	get_left_in_token(token(Left,ALeft,_)) => EV2,
	nth0(N,Left,NewEV2), nth0(N,ALeft,NewAEV2),
	re_use(EV1,AEV1,NewEV2,NewAEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN), !.
re_use(EV1,_AEV1,EV2,AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN) :- 
	get_right_in_token(token(Right,ARight,_)) => EV1,
	nth0(N,Right,NewEV1), nth0(N,ARight,NewAEV1),
	re_use(NewEV1,NewAEV1,EV2,AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN), !.
re_use(EV1,_AEV1,EV2,AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN) :- 
	get_left_in_token(token(Left,ALeft,_)) => EV1,
	nth0(N,Left,NewEV1), nth0(N,ALeft,NewAEV1),
	re_use(NewEV1,NewAEV1,EV2,AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN), !.
re_use(EV1,AEV1,EV2,_AEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN) :- 
	get_right_in_token(token(Right,ARight,_)) => EV2,
	nth0(N,Right,NewEV2), nth0(N,ARight,NewAEV2),
	re_use(EV1,AEV1,NewEV2,NewAEV2,Left_in_token,Right_in_token,CEventOID,OUTTOKEN).

go_back(Var,OtherVar,Events,AntiEvents,Left_in_token,Right_in_token,Next_op,CEventOID,Out_token) :- 
	nth0(N,Events,EV), nth0(N,AntiEvents,AEV),
	get_right_in_token(token(RL1,RL2,Right_in_templ)) => EV,	% Get the right in-token template
	get_left_in_token(token(LL1,LL2,Left_in_templ)) => EV,  	% Get the left in-token template
	((member(var(Var,Val,true,true),Left_in_templ),   % Does the Y var exist from left?
	  member(var(OtherVar,Val,true,true),Right_in_templ)) ->  % Does the X var exist on the right?
		(create_join_event(mm_positive_event,Var,EV,AEV,['='],Left_in_token,EV,AEV,['\\=='],
				   Right_in_token,CEventOID,Next_op,Out_token),		% If yes, then perform the join
		 update_right_in_token([token(RL1,RL2,Right_in_templ),     		% and store the Y var on the right 
					token(RL1,RL2,[var(Var,Val,true,true)|Right_in_templ])]) => CEventOID);
		(member(var(Var,Val,true,true),Left_in_templ) ->  % Try to go on the left
			(templ_to_token(Left_in_templ,NextLeftInToken),
			 go_back(Var,OtherVar,LL1,LL2,NextLeftInToken,Right_in_token,Next_op,CEventOID,Out_token));
			(member(var(OtherVar,Val,true,true),Right_in_templ) -> 
				(templ_to_token(Right_in_templ,NextRightInToken),
			 	 go_back(Var,OtherVar,RL1,RL2,Left_in_token,NextRightInToken,Next_op,CEventOID,Out_token));
				(write('Problem with X@class(...)\\=Y case!'), nl,
				 fail)))).
