%****************************************************************************************

%*************************************************************************
% FILE:		device_event_manager.pl
% SUBJECT:	It is the definition of metaclasses and classes that 
%		compound the event manager for the DEVICE system
% AUTHOR:	NICK BASSILIADES
% DATE:		October-December 1995
%*************************************************************************

:-

%*************************************************************************
%*************************************************************************
%*************************************************************************
%			ENTITIES: METACLASSES
%*************************************************************************
%*************************************************************************
%*************************************************************************

   new([mm_device_event_manager,[
    is_a([mm_event_manager]),

%   +--------------------------------------------------------------------+
%   |	method SIGNAL for DEVICE events					 |
% For details look at the comments of SIGNAL method in the event_manager.pl 
% file of EXACT

    method((signal(global,[],[plog],[],[SingleOccs]) :-
     (	
	% The following lines have been added for DEVICE
	% Scheduling of DEVICE rules are done after the scheduling of the rest
	% of rules and after the de_scheduling of DEVICE rules. The scheduling
        % is done through the same method of mm_generic_rule that is inherited
        % The message is however sent to mm_production_rule, just to have
        % altered calculate_conflict_set method!
	device_signal([SingleOccs],DeviceOccs) => mm_device_event,
	% The primitive events are re-joined to cater for the case that 
	% a single event evokes a production rule!
	(DeviceOccs \== [] ->
		append(SingleOccs,DeviceOccs,DeviceAllOccs);
		DeviceAllOccs = SingleOccs),
	schedule_all([DeviceAllOccs]) => mm_production_rule_manager


    ))),

%    Method specific to DEVICE
%   +--------------------------------------------------------------------+
%   |	method DEVICE_SIGNAL						 |
%   |	For every event occurrence in Occs the method does:
%   |	it looks in the event object for a DEVICE event that might be
%   |	activated by the event that occurred. It checks then if it   
%   |	is possible to add occurrences of any of these DEVICE events
%   |
%   |	It also does the same considering the new inclusions of DEVICE
%   |	event occurrences and loops in this way until no new event occ.
%   |	is found
%   |
    method((device_signal(global,[],[plog],plog,[[Occs],DeviceOccs]) :- 
     (	
	findall( AdditionalOcc,
		 (member(Occ,Occs), 
		  Occ =.. [occurrence_tuple,Event,ParamsTuple1],
		  (ParamsTuple1 = unique(ParamsTuple) -> 
			true;
			ParamsTuple = ParamsTuple1),
		  (my_instance_of(Event,mm_inheritable_ms_event) -> 
		      (
		       (setof(E,get_activated_events(E) => Event, Es),
			member(DeviceEvent,Es),
			PosNeg = positive);
		       (setof(E,get_de_activated_events(E) => Event, Es),
			member(DeviceEvent,Es),
			PosNeg = negative));
		      (
		       ParamsTuple =.. [PosNeg,Params],
		       (PosNeg == positive ->  % The event is not primitive; it may be pos or neg!
		      		get_correct_activated_events([Params],DeviceEvent) => Event;
				get_correct_de_activated_events([Params],DeviceEvent) => Event))),
		  (is_list(DeviceEvent) -> 
		      	(transform_update_parameters(DeviceEvent,DEvent,Occ,[Ptuple,Dtuple]),
			 (anti_checking([Dtuple],AdditionalOcc) => DEvent;
			  checking([Ptuple],AdditionalOcc) => DEvent));
		  	(transform_parameters(DeviceEvent,Occ,PosNeg,Occ1),
			 (PosNeg == positive -> 
		  	 	checking([Occ1],AdditionalOcc) => DeviceEvent;
				anti_checking([Occ1],AdditionalOcc) => DeviceEvent)))),
	FirstSetOccs),
	(FirstSetOccs == []
          -> DeviceOccs = FirstSetOccs
	  %  Loop until no new event occurrence is added:
	  %
          ;  (device_signal([FirstSetOccs],SecondSetOccs) => mm_device_event,
	      (SecondSetOccs \== [] -> 
	     		append(FirstSetOccs,SecondSetOccs,DeviceOccs);
			DeviceOccs = FirstSetOccs))
	)
    )))


]]) => meta_class,
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%*************************************************************************
%*************************************************************************
%*************************************************************************
%			ENTITIES: CLASSES
%*************************************************************************
%*************************************************************************
%*************************************************************************

   new([mm_device_event,[
    is_a([mm_generic_event])
   ]]) => mm_device_event_manager,




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
% This type of event is signaled by the EXECUTION OF THE METHOD 
% 'active_method' defined in the class 'active_class' (or either sent to 
% the object 'active_object'.
% The 'when' slot value indicates if the event is going to be signaled
% BEFORE method execution, AFTER, BEFORE-EACH or else if the method FAIL 
% or else if the method is NOT_FOUND.
%
   new([mm_logical_event,[
    is_a([mm_device_event]),

%   ----------------------------------------------------------------------
    slot(slot_tuple(in_token,		global,set,total,plog)),
    slot(slot_tuple(predicate,		global,single,total,plog)),


     method((get_activated_by(global, set, [], mm_ms_event, [EventOID]) :- 
	              get_in_token(token(EventOIDs,_,_)) => self,
		      member(EventOID,EventOIDs)
	)),
     method((get_de_activated_by(global, set, [], mm_ms_event, [EventOID]) :- 
	              get_in_token(token(_,EventOIDs,_)) => self,
		      member(EventOID,EventOIDs)
	)),

     method((get_all_activated_by(global, set, [], mm_ms_event, [EventOIDs]) :- 
	              findall(E,get_activated_by(E) => self,EventOIDs)

	)),
     method((get_all_de_activated_by(global, set, [], mm_ms_event, [EventOIDs]) :- 
	              findall(E,get_de_activated_by(E) => self,EventOIDs)

	)),

     method((get_event_net(global, set, [], mm_ms_event, [EventOIDs]) :- 
		      % message_recipient(MR),
	              findall(E,get_activated_by(E) => self,EventOIDs)

	)),



%   +--------------------------------------------------------------------+
%   |	method SET_ALARM						 |
%   |	It sends the 'detector_on' message preparing the arguments from
%   |	the slots values.
%   |	(for more info see the supermethod in mm_generic_event)
%   |


    method((set_alarm(global,[],[],[],[]) :-
	(
	 ((findall([Ev,AEv],(get_in_token(token(EventsIn,AntiEventsIn,_)) => self,
	                     nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
		   Event_List1),
	   remove_dups(Event_List1,Event_List),
	   member([Ev,AEv],Event_List),
	   set_alarm => Ev,
	   set_alarm => AEv,
	   find_update_event(AEv,UpdEv),
	   set_alarm => UpdEv,
	   fail) ; true)
   ))),

    method((enable_event(global,[],[object],[],[Sender]) :-
	(
	 message_recipient(LEvent),
	 ((findall([Ev,AEv],(get_in_token(token(EventsIn,AntiEventsIn,_)) => self,
	                     nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
		   Event_List1),
	   remove_dups(Event_List1,Event_List),
	   member([Ev,AEv],Event_List),
	   enable_event([LEvent]) => Ev,
	   set_alarm => AEv,
	   find_update_event(AEv,UpdEv),
	   %get(UpdEv) => mm_inheritable_ms_event,
	   %get_out_token(token([[Ev,AEv]],[],_)) => UpdEv,
	   set_alarm => UpdEv,
	   fail) ; true)
   ))),


%   +--------------------------------------------------------------------+
%   |	method REMOVE_ALARM						 |
%   |	It has just the opposite behavior than 'set_alarm'.
%   |
    method((remove_alarm(global,[],[],[],[]) :-
     (((findall([Ev,AEv],(get_in_token(token(EventsIn,AntiEventsIn,_)) => self,
	                  nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
		Event_List1),
	remove_dups(Event_List1,Event_List),
	member([Ev,AEv],Event_List),
	remove_alarm => Ev,
	remove_alarm => AEv,
	find_update_event(AEv,UpdEv),
	remove_alarm => UpdEv,
	fail) ; true)
   ))),

%   +--------------------------------------------------------------------+
%   |	method CHECKING							 |
%   |
%   |	It checks if a logical event can be raised (as it is shown  
%   |	in the class header). ActualOcc is in the form:
%   |		occurrence_tuple(EventOID, Params)      
%   |	(for more info about Params see 'signal' defined in mm_event_manager)
%   |
%   |
    method((checking(global,[],[occurrence_tuple],occurrence_tuple,[[ActualOcc],LogicalOcc]) :-
     (	message_recipient(Logical_event),
	ActualOcc = occurrence_tuple(_,positive(In_token)),
	(get_spy(on) => Logical_event ->
		write_list(['Event:',Logical_event,'Input:', ActualOcc]);
		true),
	(In_token = [var(Var,[Val])];
	 In_token = [var(OID_Var,[OID]),var(Var,[Val])];
	 In_token = [var(OID_Var,[OID]),var(Var1,[Val]),var(Var2,[Val])]),
	get_predicate([Val,Pred]) => Logical_event, !,
	(call(Pred) ->
		 (get_out_token(token(_,_,Out_token)) => Logical_event,
	  	  my_subset(Out_token,In_token),
	  	  LogicalOcc = occurrence_tuple(Logical_event,positive(Out_token)),
	  	  (get_spy(on) => Logical_event ->
			write_list(['Event:',Logical_event,'Output:',LogicalOcc]), nl;
			true));
		 (get_spy(on) => Logical_event ->
			write_list(['Event:',Logical_event,'No Output!']), nl, fail;
			fail))
   ))),

%   +--------------------------------------------------------------------+
%   |	method ANTI_CHECKING						 |
%   |
%   |	It checks if a logical anti-event can be raised. 
%   |		occurrence_tuple(EventOID, Params)      
%   |
%   |
    method((anti_checking(global,[],[occurrence_tuple],occurrence_tuple,[[ActualOcc],LogicalOcc]) :-
     (	message_recipient(Logical_event),
	ActualOcc = occurrence_tuple(_,negative(In_token)),
	(get_spy(on) => Logical_event ->
		write_list(['Event:',Logical_event,'Input:', ActualOcc]);
		true),
	(In_token = [var(Var,[Val])];
	 In_token = [var(OID_Var,[OID]),var(Var,[Val])];
	 In_token = [var(OID_Var,[OID]),var(Var1,[Val]),var(Var2,[Val])]),
	get_predicate([Val,Pred]) => Logical_event, !,
	(call(Pred) ->
		 (get_out_token(token(_,_,Out_token)) => Logical_event,
	  	  my_subset(Out_token,In_token),
	  	  LogicalOcc = occurrence_tuple(Logical_event,negative(Out_token)),
		  (get_spy(on) => Logical_event ->
			write_list(['Event:',Logical_event,'Output:',LogicalOcc]), nl;
			true));
		 (get_spy(on) => Logical_event ->
			write_list(['Event:',Logical_event,'No Output!']), nl, fail;
			fail))
   ))),

     method((retrieve_out_token(global, set, [], plog, [Out_token]) :- 
     	 (
	  message_recipient(LogicalEvent),
	  get_activated_by(Ev) => LogicalEvent,
	  retrieve_out_token(Token) => Ev,
	  checking([occurrence_tuple(_,positive(Token))],occurrence_tuple(_,positive(Out_token))) => LogicalEvent
    )))


   ]]) => mm_device_event_manager,
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
% DEVICE Events that are conjunction of 2 other events (single or composite)
%
% This is an abstract class, i.e. it only holds the structure for the 
% positive and negative DEVICE events, whose behaviour (methods) are different.
% (see below)

   new([mm_two_input_event,[
    is_a([mm_device_event]),

    slot(slot_tuple(join,			global,single,optional,plog)),
    slot(slot_tuple(left_in_token,		global,set,total,plog)),
    slot(slot_tuple(right_in_token,		global,set,total,plog)),
    slot(slot_tuple(left_memory,		global,set,optional,plog)),
    slot(slot_tuple(left_virtual,		global,single,total,plog)),
    slot(slot_tuple(left_counter,		global,single,total,string)),
    slot(slot_tuple(right_memory,		global,set,optional,plog)),
    slot(slot_tuple(right_virtual,		global,single,total,plog)),
    slot(slot_tuple(right_counter,		global,single,total,string)),
    slot(slot_tuple(unique_output,		global,single,optional,string)),
    slot(slot_tuple(is_enabled,			family,single,mandatory,string)),

    % This slot is only needed for deductive rules. If set to off, anti-achecking will perform
    % joins at the positive events. For negative events this is not needed, because the negative input
    % does not contribute to the output tokens anyway!
    slot(slot_tuple(deletion_optimization,	global,single,total,string)),


     method((get_left_activated_by(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_left_in_token(token(EventOIDs,_,_)) => self,
		      member(EventOID,EventOIDs)
	)),
     method((get_left_de_activated_by(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_left_in_token(token(_,EventOIDs,_)) => self,
		      member(EventOID,EventOIDs)
	)),
     method((get_right_activated_by(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_right_in_token(token(EventOIDs,_,_)) => self,
		      member(EventOID,EventOIDs)
	)),
     method((get_right_de_activated_by(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_right_in_token(token(_,EventOIDs,_)) => self,
		      member(EventOID,EventOIDs)
	)),

     method((get_all_activated_by(global, single, [], mm_generic_event, [EventOIDs]) :- 
	              findall(E,(get_left_activated_by(E) => self;
		                 get_right_activated_by(E) => self),
			      FirstSet),
		      findall(L,(member(X,FirstSet),
		                 get_all_activated_by(L) => X),
			      Lists),
		      append_lists([FirstSet|Lists],EventOIDs)
	)),
     method((get_all_de_activated_by(global, single, [], mm_generic_event, [EventOIDs]) :- 
	              findall(E,(get_left_de_activated_by(E) => self;
		                 get_right_de_activated_by(E) => self),
			      FirstSet),
		      findall(L,(member(X,FirstSet),
		                 get_all_de_activated_by(L) => X),
			      Lists),
		      append_lists([FirstSet|Lists],EventOIDs)
	)),

     method((get_event_net(global, single, [], mm_generic_event, [Event_list]) :- 
		      % message_recipient(MR),
	              findall(LL,(get_left_activated_by(E) => self,
				  (LL = E;
				   (get_event_net(N) => E -> 
					LL = N))),
			      LeftLists),
	              findall(LL,(get_right_activated_by(E) => self,
				  (LL = E;
				   (get_event_net(N) => E -> 
					LL = N))),
			      RightLists),
		      append(LeftLists,RightLists,Event_list)
	)),


%   +--------------------------------------------------------------------+
%   |	method SET_ALARM						 |
%   |	It sends the 'set_alarm' method to each component event
%   |	DUDA:  LOS EVENTOS SIMPLES DE MENSAJE COMPONENTES QUE YA HAN RECIBIDO  	?????
%   |	UN DETECTOR_ON (SE HA HECHO UN STORE), MEDIANTE LA EJECUCION DE   	?????
%   | 	ESTE MENSAJE VAN A RECIBIR OTRO STORE.  VA A DEJAR LAS COSAS BIEN? 	?????
%   | 	UN REMOVE DEL COMPOSITE VA A BORRAR EL SINGLE? UN REMOVE DEL SINGLE 	?????
%   | 	VA A DEJAR BIEN EL COMPOSITE???
%   |	(for more info see the supermethod in mm_generic_event)
%   |

    method((set_alarm(global,[],[],[],[]) :-
     (
	message_recipient(Event),
      	((findall([Ev,AEv],(get_right_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		  	    nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
		  Event_list1),
	  remove_dups(Event_list1,Event_list),
	  member([Ev,AEv],Event_list),
	  set_alarm => Ev,
	  (Ev == AEv -> 
		true;
		(set_alarm => AEv,
		 find_update_event(AEv,UpdEv),
		 set_alarm => UpdEv)),
	  fail) ; true),
	((findall([Ev,AEv],(get_left_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		   	    nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
		  Event_list1),
	  remove_dups(Event_list1,Event_list),
	  member([Ev,AEv],Event_list),
	  set_alarm => Ev,
	  (Ev == AEv -> 
		true;
		(set_alarm => AEv,
		 find_update_event(AEv,UpdEv),
		 set_alarm => UpdEv)),
	  fail) ; true)

   ))),


    method((enable_event(global,[],[object],[],[Sender]) :-
     (
	message_recipient(Event),
	(get_is_enabled(on) => Event ->
	 	 (my_instance_of(Sender,mm_production_rule) -> 
			(findall(Occ,(retrieve_out_token(Token) => Event,
				      Occ = occurrence_tuple(Event,positive(Token))),
				 AllOccs),
			 schedule_all([AllOccs]) => mm_production_rule_manager);	
	 		(findall(NextOcc,(retrieve_out_token(Token) => Event,
				      	  Occ = occurrence_tuple(Event,positive(Token)),
				          checking([Occ],NextOcc) => Sender),
				 AllOccs),
			 signal([AllOccs]) => mm_device_event));
      		((get_is_enabled(off) => Event ->
		    update_is_enabled([off,on]) => Event;
		    put_is_enabled([on]) => Event),
		 get_left_virtual(virtual(VirtualL,_)) => Event,
		 get_right_virtual(virtual(VirtualR,_)) => Event,
		% The right input is enabled first, for the sake of negative events
		% For positive events is all the same!
      		((findall([Ev,AEv],(get_right_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		  		    nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
			   Event_list1),
		  remove_dups(Event_list1,Event_list),
		  member([Ev,AEv],Event_list),
		  (VirtualR == false ->
 	  		enable_event([Event]) => Ev;
			set_alarm => Ev),
		  (Ev == AEv -> 
			true;
			(set_alarm => AEv,
			 find_update_event(AEv,UpdEv),
		 	 %get(UpdEv) => mm_inheritable_ms_event,
		 	 %get_out_token(token([[Ev,AEv]],[],_)) => UpdEv,
		 	 set_alarm => UpdEv)),
	 	  fail) ; true),
		 ((findall([Ev,AEv],(get_left_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		   		     nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
			   Event_list1),
		   remove_dups(Event_list1,Event_list),
		   member([Ev,AEv],Event_list),
		   ((VirtualL == false; (VirtualL == true, VirtualR == true)) ->
 	  		enable_event([Event]) => Ev;
			set_alarm => Ev),
		   (Ev == AEv -> 
			true;
			(set_alarm => AEv,
			 find_update_event(AEv,UpdEv),
		 	 %get(UpdEv) => mm_inheritable_ms_event,
		 	 %get_out_token(token([[Ev,AEv]],[],_)) => UpdEv,
		 	 set_alarm => UpdEv)),
	 	  fail) ; true)
		))
   ))),


%   +--------------------------------------------------------------------+
%   |	method REMOVE_ALARM						 |
%   |	It has just the opposite behavior than 'set_alarm'.
%   |
    method((remove_alarm(global,[],[],[],[]) :-
     (%%,
	(get_is_enabled(on) => self ->
      		(((findall([Ev,AEv],(get_left_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		   		     nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
			   Event_list1),
		   remove_dups(Event_list1,Event_list),
		   member([Ev,AEv],Event_list),
		  (Ev == AEv -> 
			remove_alarm => Ev;
			(remove_alarm => Ev,
		 	 remove_alarm => AEv,
			 find_update_event(AEv,UpdEv),
		 	 remove_alarm => UpdEv)),
	 	 fail) ; true),
      		 ((get_left_memory(Token) => self,
		  delete_left_memory([Token]) => self,
		  fail); true),
      		((findall([Ev,AEv],(get_right_in_token(token(EventsIn,AntiEventsIn,_)) => self,
		  		    nth0(I,EventsIn,Ev), nth0(I,AntiEventsIn,AEv)),
			   Event_list1),
		  remove_dups(Event_list1,Event_list),
		  member([Ev,AEv],Event_list),
		  (Ev == AEv -> 
			remove_alarm => Ev;
			(remove_alarm => Ev,
		 	 remove_alarm => AEv,
			 find_update_event(AEv,UpdEv),
		 	 remove_alarm => UpdEv)),
	 	 fail) ; true),
      		((get_right_memory(Token) => self,
		  delete_right_memory([Token]) => self,
		  fail); true),
		update_is_enabled([on,off]) => self);
		true)
   ))),


    method((insert_left(global,[],[plog],[],[In_token]) :-
	(	
		message_recipient(CompositeEvent),
		  get_left_virtual(virtual(VirtualL,_)) => CompositeEvent,
		  get_left_counter(LeftCounter) => CompositeEvent,
		  (VirtualL == false -> 
			(LeftCounter == on ->
		 	 	(get_left_memory(token(In_token,Count)) => CompositeEvent -> 
		  			(New_Count is Count + 1,
			 	 	 update_left_memory([token(In_token,Count),token(In_token,New_Count)]) => CompositeEvent,
			 	 	 !, fail);   % If Count>1 the token pre-exists; Fail and don't return any output tokens!
					put_left_memory([token(In_token,1)]) => CompositeEvent);
				put_left_memory([token(In_token)]) => CompositeEvent);
			true)
    ))),


    method((insert_right(global,[],[plog],[],[In_token]) :-
	(	
		message_recipient(CompositeEvent),
		 get_right_virtual(virtual(VirtualR,_)) => CompositeEvent,
		 get_right_counter(RightCounter) => CompositeEvent,
		 (VirtualR == false -> 
			(RightCounter == on ->
		 		(get_right_memory(token(In_token,Count)) => CompositeEvent -> 
		  			(New_Count is Count + 1,
			 	 	 update_right_memory([token(In_token,Count),token(In_token,New_Count)]) => CompositeEvent,
			 	 	 !, fail);   % If Count>1 the token pre-exists; Fail and don't return any output tokens!
					put_right_memory([token(In_token,1)]) => CompositeEvent);
				put_right_memory([token(In_token)]) => CompositeEvent);
			true)
    ))),


    method((remove_left(global,[],[plog],[],[In_token]) :-
	(	
		message_recipient(CompositeEvent),
		get_left_virtual(virtual(VirtualL,_)) => CompositeEvent,
		get_left_counter(LeftCounter) => CompositeEvent,
		(VirtualL == false -> 
			(LeftCounter == on ->
			 	(get_left_memory(token(In_token,Count)) => CompositeEvent,
			  	 (Count > 1 ->
					(NewCount is Count - 1,
				 	 update_left_memory([token(In_token,Count),
							     token(In_token,NewCount)]) => CompositeEvent,
				 	 !, fail);
					delete_left_memory([token(In_token,1)]) => CompositeEvent));
				(findall(IT,get_left_memory(token(IT)) => CompositeEvent,IT1s),
				 remove_dups(IT1s,ITs),
				 (ITs \== [] ->
					(\+ member_set(In_token,ITs) -> 
						true;
						(member_set(In_token,ITs),
					 	 delete_left_memory([token(In_token)]) => CompositeEvent));
					fail)));
			true)
    ))),


    method((remove_right(global,[],[plog],[],[In_token]) :-
	(	
		message_recipient(CompositeEvent),
		get_right_virtual(virtual(VirtualR,_)) => CompositeEvent,
		get_right_counter(RightCounter) => CompositeEvent,
		(VirtualR == false -> 
			(RightCounter == on ->
			 	(get_right_memory(token(In_token,Count)) => CompositeEvent,
			  	 (Count > 1 ->
					(NewCount is Count - 1,
				 	 update_right_memory([token(In_token,Count),
							      token(In_token,NewCount)]) => CompositeEvent,
				 	 !, fail);
					delete_right_memory([token(In_token,1)]) => CompositeEvent));
				(findall(IT,get_right_memory(token(IT)) => CompositeEvent,IT1s),
				 remove_dups(IT1s,ITs),
				 (ITs \== [] ->
					(\+ member_set(In_token,ITs) ->
						true;
						(member_set(In_token,ITs),
					 	 delete_right_memory([token(In_token)]) => CompositeEvent));
					fail)));
			true)
    ))),



    method((join_lr(global,[],[plog],plog,[[In_token],OUT]) :-
	(	
		message_recipient(CompositeEvent),
		(get_join(Join_var_list) => CompositeEvent ->
			true;
			Join_var_list = []),
		collect_join_vars(lr,Join_var_list,Predicates),
		get_right_virtual(virtual(VirtualR,GoalList)) => CompositeEvent,
		get_right_counter(RightCounter) => CompositeEvent,
		(VirtualR == false -> 
		 	 (get_right_in_token(token(_,_,STokenTemplate)) => CompositeEvent,
			  templ_to_token(STokenTemplate,SToken),
			  partially_bound(In_token,SToken,Join_var_list,Predicates), !,
			  (RightCounter == on ->
			  	get_right_memory(token(SToken,_)) => CompositeEvent;
			  	get_right_memory(token(SToken)) => CompositeEvent));
			 (choose_goal(In_token,SToken,Join_var_list,Predicates,GoalList,Call), !,
			  call(Call))),
		check_all(lr,Join_var_list,In_token,SToken,OUT)
    ))),

    method((join_rl(global,[],[plog],plog,[[In_token],OUT]) :-
	(	message_recipient(CompositeEvent),
		(get_join(Join_var_list) => CompositeEvent ->
			true;
			Join_var_list = []),
		collect_join_vars(rl,Join_var_list,Predicates),
		get_left_virtual(virtual(VirtualL,GoalList)) => CompositeEvent,
		get_left_counter(LeftCounter) => CompositeEvent,
		(VirtualL == false -> 
		 	(get_left_in_token(token(_,_,STokenTempl)) => CompositeEvent,
			 templ_to_token(STokenTempl,SToken),
			 partially_bound(In_token,SToken,Join_var_list,Predicates), !,
			 (LeftCounter == on ->
			 	get_left_memory(token(SToken,_)) => CompositeEvent;
			 	get_left_memory(token(SToken)) => CompositeEvent));
			(choose_goal(In_token,SToken,Join_var_list,Predicates,GoalList,Call), !,
			 call(Call))),
		check_all(rl,Join_var_list,SToken,In_token,OUT)
    ))),

    method((merge_lr(global,[],[plog],plog,[[In_token],OUT]) :-
	(	
		message_recipient(CompositeEvent),
		(get_join(Join_var_list) => CompositeEvent ->
			true;
			Join_var_list = []),
		get_right_in_token(token(_,_,STokenTemplate)) => CompositeEvent,
		templ_to_token(STokenTemplate,SToken),
		merge_all(lr,Join_var_list,In_token,SToken,OUT)
    ))),

    method((merge_rl(global,[],[plog],plog,[[In_token],OUT]) :-
	(	
		message_recipient(CompositeEvent),
		(get_join(Join_var_list) => CompositeEvent ->
			true;
			Join_var_list = []),
		get_left_in_token(token(_,_,STokenTemplate)) => CompositeEvent,
		templ_to_token(STokenTemplate,SToken),
		merge_all(rl,Join_var_list,SToken,In_token,OUT)
    )))


   ]]) => mm_device_event_manager,
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   replace_method([
	(get_left_memory(global,set,[],plog,[TokenTerm]) :- 
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_left_in_token(token(_,_,TokenTemplate)) => self,
			 (nonvar(MemoryToken) -> 
			 	retrieve_vars(MemoryToken,TokenTemplate,StoredToken,RetrieveCalls);
				(collect_vars(MemoryToken1,TokenTemplate,StoredToken,RetrieveCalls),
				 remove_dups(MemoryToken1,MemoryToken))),
			 TokenTerm1 =.. [token,StoredToken|Rest], 
			 !,
			 TokenTerm1 <= left_memory,
			 call(RetrieveCalls));
			TokenTerm <= left_memory))
	]) => mm_two_input_event,

   replace_method([
	(get_right_memory(global,set,[],plog,[TokenTerm]) :- 
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_right_in_token(token(_,_,TokenTemplate)) => self,
			 (nonvar(MemoryToken) -> 
			 	retrieve_vars(MemoryToken,TokenTemplate,StoredToken,RetrieveCalls);
				(collect_vars(MemoryToken1,TokenTemplate,StoredToken,RetrieveCalls),
				 remove_dups(MemoryToken1,MemoryToken))),
			 TokenTerm1 =.. [token,StoredToken|Rest], 
			 !,
			 TokenTerm1 <= right_memory,
			 call(RetrieveCalls));
			TokenTerm <= right_memory))
	]) => mm_two_input_event,

  replace_method([
 	(put_left_memory(global,[],[plog],[],[TokenTerm]) :- 
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_left_in_token(token(_,_,TokenTemplate)) => self,
			 retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
			 TokenTerm1 =.. [token,StoredToken|Rest],
			 left_memory <-- TokenTerm1);
			left_memory <-- TokenTerm)) % This is done to evoke the ADAM error message!
	]) => mm_two_input_event,

  replace_method([
 	(put_right_memory(global,[],[plog],[],[TokenTerm]) :- 
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_right_in_token(token(_,_,TokenTemplate)) => self,
			 retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
			 TokenTerm1 =.. [token,StoredToken|Rest],
			 right_memory <-- TokenTerm1);
			right_memory <-- TokenTerm)) % This is done to evoke the ADAM error message!
	]) => mm_two_input_event,

  replace_method([
 	(delete_left_memory(global,[],[plog],[],[TokenTerm]) :-
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_left_in_token(token(_,_,TokenTemplate)) => self,
			 retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
			 TokenTerm1 =.. [token,StoredToken|Rest],
			 left_memory <-- TokenTerm1/[]);
			left_memory <-- TokenTerm/[])) % This is done to evoke the ADAM error message!
	]) => mm_two_input_event,

  replace_method([
 	(delete_right_memory(global,[],[plog],[],[TokenTerm]) :-
		(nonvar(TokenTerm) ->
			(TokenTerm =.. [token,MemoryToken|Rest],
			 get_right_in_token(token(_,_,TokenTemplate)) => self,
			 retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
			 TokenTerm1 =.. [token,StoredToken|Rest],
			 right_memory <-- TokenTerm1/[]);
			right_memory <-- TokenTerm/[])) % This is done to evoke the ADAM error message!
	]) => mm_two_input_event,

  replace_method([
 	(update_left_memory(global,[],[plog,plog],[],[token(MemoryToken,OldCount),
						 token(MemoryToken,NewCount)]) :- 
		get_left_in_token(token(_,_,TokenTemplate)) => self,
		retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
		left_memory <-- token(StoredToken,OldCount)/token(StoredToken,NewCount))
	]) => mm_two_input_event,

  replace_method([
 	(update_right_memory(global,[],[plog,plog],[],[token(MemoryToken,OldCount),
						  token(MemoryToken,NewCount)]) :- 
		get_right_in_token(token(_,_,TokenTemplate)) => self,
		retrieve_stored_vars(MemoryToken,TokenTemplate,StoredToken),
		right_memory <-- token(StoredToken,OldCount)/token(StoredToken,NewCount))
	]) => mm_two_input_event,


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
%  DEVICE positive events
%
   new([mm_positive_event,[
    is_a([mm_two_input_event]),


%   +--------------------------------------------------------------------+
%   |	method CHECKING							 |
%   |	It checks if a sequence event can be raised (as it is shown in 
%   |	th%e class header). ActualOcc is in the form:
%   |		occurrence_tuple(EventOID, Params)      
%   |	(for more info about Params see 'signal' defined in mm_event_manager)
%   |
%   |	Notice that checking fails if no composite event is found
%   |
    method((checking(global,[],[occurrence_tuple],occurrence_tuple,
            [[ActualOcc],CompositeOcc]) :-
     (	
	message_recipient(CompositeEvent), 

	(ActualOcc = occurrence_tuple(ConstituentEvent,unique(positive(In_token))) -> 
		Unique = true;
		ActualOcc = occurrence_tuple(ConstituentEvent,positive(In_token)),
		Unique = false),
	(get_left_activated_by(ConstituentEvent) => CompositeEvent -> 
		( 
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Left-Input:',ActualOcc]);
			true),
		  get_left_in_token(token(_,_,Left_Templ)) => CompositeEvent,
		  templ_to_token(Left_Templ,Left_in_token),
		  my_subset(In_token,Left_in_token),
		  (Unique = true ->
			findall(NCC,(remove_left([AStoredToken]) => CompositeEvent,
				      (\+ get_activated_rules(_) => CompositeEvent -> 
			 			merge_lr([AStoredToken],OUT) => CompositeEvent;
						join_lr([AStoredToken],OUT) => CompositeEvent),
		   	      	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	              same_set(Out_token,OUT),
		   	              NCC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			        NCCs);
			NCCs = []),
		  insert_left([Left_in_token]) => CompositeEvent,
		  findall(PCC,(join_lr([Left_in_token],OUT) => CompositeEvent,
		    	       get_out_token(token(_,_,Out_token)) => CompositeEvent,
	            	       same_set(Out_token,OUT),
			       (get_unique_output(yes) => CompositeEvent ->
					PCC = occurrence_tuple(CompositeEvent,unique(positive(Out_token)));
	            	       		PCC = occurrence_tuple(CompositeEvent,positive(Out_token)))),
	            	  PCCs),	          
		  append(NCCs,PCCs,CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail)));
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Right-Input:',ActualOcc]);
			true), 
		  get_right_in_token(token(_,_,Right_Templ)) => CompositeEvent,
		  templ_to_token(Right_Templ,Right_in_token),
		  my_subset(In_token,Right_in_token),
		  (Unique = true ->
			findall(NCC,(remove_right([AStoredToken]) => CompositeEvent,
				      (\+ get_activated_rules(_) => CompositeEvent -> 
			 			merge_rl([AStoredToken],OUT) => CompositeEvent;
						join_rl([AStoredToken],OUT) => CompositeEvent),
		   	      	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	              same_set(Out_token,OUT),
		   	              NCC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			        NCCs);
			NCCs = []),
		  insert_right([Right_in_token]) => CompositeEvent,		 
		  findall(PCC,(join_rl([Right_in_token],OUT) => CompositeEvent,
		   	       get_out_token(token(_,_,Out_token)) => CompositeEvent,
	           	       same_set(Out_token,OUT),
			       (get_unique_output(yes) => CompositeEvent ->
					PCC = occurrence_tuple(CompositeEvent,unique(positive(Out_token)));
	            	       		PCC = occurrence_tuple(CompositeEvent,positive(Out_token)))),
			  PCCs),
		  append(NCCs,PCCs,CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail))))

    ))),




%   +--------------------------------------------------------------------+
%   |	method ANTI_CHECKING						 |
%   |	It checks if a positive two-input anti-event can be raised  
%   |
%   |
    method((anti_checking(global,[],[occurrence_tuple],occurrence_tuple,
            [[ActualOcc],CompositeOcc]) :-
     (
	message_recipient(CompositeEvent),

	% The unique keyword does not affect anti-checking!
	(ActualOcc = occurrence_tuple(ConstituentEvent,unique(negative(In_token))) ->
		true;
		ActualOcc = occurrence_tuple(ConstituentEvent,negative(In_token))),
	(get_left_de_activated_by(ConstituentEvent) => CompositeEvent -> 
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Left-Input:',ActualOcc]);
			true),
		  get_left_in_token(token(_,_,Left_Templ)) => CompositeEvent,
		  templ_to_token(Left_Templ,Left_in_token),
		  my_subset(In_token,Left_in_token),
		  findall(CC,(remove_left([Left_in_token]) => CompositeEvent,
		              % Delete-optimization is off only if the current event outputs to a rule!
			      % or when the join is not equi-join
		              ((\+ get_activated_rules(_) => CompositeEvent,
				get_deletion_optimization(on) => CompositeEvent) -> 
			 		merge_lr([Left_in_token],OUT) => CompositeEvent;
					join_lr([Left_in_token],OUT) => CompositeEvent),
		   	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	      same_set(Out_token,OUT),
		   	      CC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			  CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail)));
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Right-Input:',ActualOcc]);
			true),
		  get_right_in_token(token(_,_,Right_Templ)) => CompositeEvent,
		  templ_to_token(Right_Templ,Right_in_token),
		  my_subset(In_token,Right_in_token),
		  findall(CC,(remove_right([Right_in_token]) => CompositeEvent,
		   	      ((\+ get_activated_rules(_) => CompositeEvent,
				get_deletion_optimization(on) => CompositeEvent) -> 
					merge_rl([Right_in_token],OUT) => CompositeEvent;
					join_rl([Right_in_token],OUT) => CompositeEvent),
		   	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	      same_set(Out_token,OUT),
		   	      CC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			  CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail))))

    ))),



     method((retrieve_out_token(global, set, [], plog, [Token]) :- 
	 message_recipient(CompositeEvent),
	 get_left_counter(LeftCounter) => CompositeEvent,
	 (LeftCounter == on ->
		get_left_memory(token(LToken,_)) => CompositeEvent;
		get_left_memory(token(LToken)) => CompositeEvent),
	 join_lr([LToken],OUT) => CompositeEvent,
	 get_out_token(token(_,_,Token)) => CompositeEvent,
	 same_set(Token,OUT)
    ))



   ]]) => mm_device_event_manager,





%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
%  DEVICE negative events
%
   new([mm_negative_event ,[
    is_a([mm_two_input_event]),
   

%   +--------------------------------------------------------------------+
%   |	method CHECKING							 |
%   |	It checks if a sequence event can be raised (as it is shown in 
%   |	the class header). ActualOcc is in the form:
%   |		occurrence_tuple(EventOID, Params)      
%   |	(for more info about Params see 'signal' defined in mm_event_manager)
%   |
%   |	Notice that checking fails if no composite event is found
%   |
    method((checking(global,[],[occurrence_tuple],occurrence_tuple,
            [[ActualOcc],CompositeOcc]) :-
     (	
	message_recipient(CompositeEvent),

	(ActualOcc = occurrence_tuple(ConstituentEvent,unique(positive(In_token))) -> 
		Unique = true;
		ActualOcc = occurrence_tuple(ConstituentEvent,positive(In_token)),
		Unique = false),
	(get_left_activated_by(ConstituentEvent) => CompositeEvent -> 
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Left-Input:',ActualOcc]);
			true),
		  (Unique = true ->
		 	findall(NCC,(remove_left([AStoredToken]) => CompositeEvent,
		   	     	     get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	     	     same_set(Out_token,AStoredToken),
		   	     	     NCC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			        NCCs);
			NCCs = []),
		 insert_left([In_token]) => CompositeEvent,
		 findall(PCC,((join_lr([In_token],_) => CompositeEvent ->   % Check for joins
	        		fail; 		   % If one join exists then fail
		 		true),	   	   % If all tokens are checked and none is verified then succeed.
		   	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		              same_set(Out_token,In_token),
			      (get_unique_output(yes) => CompositeEvent ->
					PCC = occurrence_tuple(CompositeEvent,unique(positive(Out_token)));
	            	       		PCC = occurrence_tuple(CompositeEvent,positive(Out_token)))),
			 PCCs),
		  append(NCCs,PCCs,CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail)));
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Right-Input:',ActualOcc]);
			true),
		 insert_right([In_token]) => CompositeEvent,
		 findall(CC,(join_rl([In_token],OUT) => CompositeEvent,
		   	     get_out_token(token(_,_,Out_token)) => CompositeEvent,
	           	     my_subset(Out_token,OUT),
			     % Check for deletion optimization!
		   	     ((get_activated_rules(_) => CompositeEvent;
			       get_deletion_optimization(off) => CompositeEvent) -> 
		 			((join_lr([Out_token],OUT2) => CompositeEvent,
			  		  \+ my_subset(In_token,OUT2)) ->   % Check for joins
	        				fail; 	% If one join exists then fail
						% If all tokens are checked and none is verified then succeed.
		 				true);	
					true),
	           	     CC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			 CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail))))


    ))),


%   +--------------------------------------------------------------------+
%   |	method ANTI_CHECKING						 |
%   |	It checks if a negative two-input anti-event can be raised  
%   |
%   |
    method((anti_checking(global,[],[occurrence_tuple],occurrence_tuple,
            [[ActualOcc],CompositeOcc]) :-
     (  
	message_recipient(CompositeEvent), 

	ActualOcc = occurrence_tuple(ConstituentEvent,negative(In_token)),
	(get_left_de_activated_by(ConstituentEvent) => CompositeEvent -> 
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Left-Input:',ActualOcc]);
			true),
		 findall(CC,(remove_left([In_token]) => CompositeEvent,
		   	     get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	     same_set(Out_token,In_token),
		   	     CC = occurrence_tuple(CompositeEvent,negative(Out_token))),
			 CCs), 
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail)));
		(
		 (get_spy(on) => CompositeEvent ->
			write_list(['Event:',CompositeEvent,'Right-Input:',ActualOcc]);
			true),
		  findall(CC,(remove_right([In_token]) => CompositeEvent,
		   	      join_rl([In_token],OUT) => CompositeEvent,
		   	      get_out_token(token(_,_,Out_token)) => CompositeEvent,
		   	      my_subset(Out_token,OUT),   % The output token is the same as the left token!
		   	      (join_lr([Out_token],_) => CompositeEvent ->   	% Check for joins
	        	  		fail; 	% If one join exists then fail
		 	  		true),	% If all tokens are checked and none is verified then succeed.
		   	      CC = occurrence_tuple(CompositeEvent,positive(Out_token))),
			 CCs),
		  (CCs \== [] ->
			 (member(CompositeOcc,CCs),
			  (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'Output:',CompositeOcc]), nl;
				true));
			 (get_spy(on) => CompositeEvent ->
				write_list(['Event:',CompositeEvent,'No Output!']), nl, fail;
				fail))))

   ))),


     method((retrieve_out_token(global, set, [], plog, [Token]) :- 
	 message_recipient(CompositeEvent),
	 get_left_counter(LeftCounter) => CompositeEvent,
	 (LeftCounter == on ->
		get_left_memory(token(LToken,_)) => CompositeEvent;
		get_left_memory(token(LToken)) => CompositeEvent),
	 (join_lr([LToken],_) => CompositeEvent ->
		fail;
		true),			 
	 get_out_token(token(_,_,Token)) => CompositeEvent,
	 same_set(Token,LToken)
    ))



]]) => mm_device_event_manager.
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
