/* exact2device.pl */

/* This file contains all the calls to alter certain
   EXACT methods, or the structure of certain EXACT classes
   and meta-classes to cope with the extended functionality
   of DEVICE. Note that all the features of EXACT can be
   used orthogonal to the features of DEVICE.

   Much of the code in here is copied from EXACT source files and 
   slightly augmented/altered. I omitted many of the comments where
   they were not mine. The original comments can be found in the
   original EXACT source files.
*/


/* The following call redefines the type checking of ADAM for the type string, which
   is extended to ECLiPSe strings as well */

:-


/* The following calls re-define structures and methods found in the event_manager.pl
   file of EXACT, and therefore have to do with the default event management of EXACT
*/



% Method 'signal' is augmented with the following active rule to cater for the DEVICE events
% and production rules, AFTER the EXACT events and ECA rules.
replace_method([
   (signal(global,[],[plog],[],[SingleOccs]) :-
     (	
	composite_signal([SingleOccs],CompositeOccs) => self,
	append(SingleOccs,CompositeOccs,TotalOccs),
	scheduling([TotalOccs]) => mm_generic_rule,
	% This is a new 'signal' method for DEVICE events
	signal([SingleOccs]) => mm_device_event, !
    ))
]) => mm_event_manager.


:-

% The slots below have been added because it is important to have
% forward references to activated events to be able to detect which events should
% be notified. It is far more efficient than the mm_composite_event method of EXACT

    put_slot([slot_tuple(out_token,global,set,optional,plog)]) => mm_generic_event,
    put_slot([slot_tuple(spy,global,single,optional,string)]) => mm_generic_event,

    put_method([(spy(global, set, [], [], []) :- 
	message_recipient(Event),
	(get_spy(A) => Event ->
	 	(A \== on ->
			(update_spy([A,on]) => Event,
		 	 write_list(['Now spying the event',Event]));
			write_list(['The event',Event,'is already being spied!']));
		(put_spy([on]) => Event,
		 write_list(['Now spying the event',Event])))
	)]) => mm_generic_event,

    put_method([(no_spy(global, set, [], [], []) :- 
	message_recipient(Event),
	(get_spy(A) => Event ->
	 	(A \== off ->
			(update_spy([A,off]) => Event,
		 	 write_list(['The event',Event,'is no more spied.']));
			write_list(['The event',Event,'is not being spied!']));
		(put_spy([off]) => Event,
		 write_list(['The event',Event,'is not spied.'])))
	)]) => mm_generic_event,

    put_method([(get_activated_events(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_out_token(token(EventOIDs,_,_)) => self,
		      member(EventOID,EventOIDs),
		      ( \+ (is_list(EventOID)) -> 
		      		\+ (my_instance_of(EventOID,mm_production_rule));
				true)
	)]) => mm_generic_event,


    put_method([(get_correct_activated_events(global, set, [plog], mm_generic_event, [[Token],EventOID]) :- 
	              get_out_token(token(EventOIDs,_,OutToken)) => self,
		      subsumes_chk(OutToken,Token),
		      member(EventOID,EventOIDs),
		      ( \+ is_list(EventOID) -> 
		      		\+ my_instance_of(EventOID,mm_production_rule);
				true)
	)]) => mm_generic_event,

    put_method([(get_de_activated_events(global, set, [], mm_generic_event, [EventOID]) :- 
	              get_out_token(token(_,EventOIDs,_)) => self,
		      member(EventOID,EventOIDs),
		      \+ (my_instance_of(EventOID,mm_production_rule))
	)]) => mm_generic_event,

    put_method([(get_correct_de_activated_events(global, set, [plog], mm_generic_event, [[Token],EventOID]) :- 
	              get_out_token(token(_,EventOIDs,OutToken)) => self,
			subsumes_chk(OutToken,Token),
		      member(EventOID,EventOIDs),
		      \+ (my_instance_of(EventOID,mm_production_rule))
	)]) => mm_generic_event,


% The following 3 (three) methods are added to override the default methods
% for the newly added  slot 'out_token' in  class mm_generic_event.
replace_method([
 (put_out_token(global,[],[plog],[],[Token]) :-
   (	message_recipient(Event),

	% If the EVENT did not have any activated events or rules attached to it previously
	% then, SET_ALARM is sent for this event
	%
	( (get_activated_events(_) => Event;
	   get_de_activated_events(_) => Event;
	   get_activated_rules(_) => Event;
	   Token=token([],[],_) )  % All tokens are 'empty'!
	  -> true 
	  ;  set_alarm => Event
	),
	out_token <-- Token
 ))
]) => mm_generic_event,

replace_method([
 (delete_out_token(global,[],[plog],[],[Token]) :-
   (	message_recipient(Event),
	out_token <-- Token/[],
	% If the EVENT does not have any activated rule or event left
	% then, REMOVE_ALARM is sent for this event
	%
	( ( (get(CompositeEvent)    => mm_composite_event, 
	     get_composed_by(Event) => CompositeEvent,
	     get_activated_rules(_) => CompositeEvent)
	    ;
	    get_activated_rules(_) => Event;
	    get_de_activated_rules(_) => Event;
	    get_activated_events(_) => Event;
	    get_de_activated_events(_) => Event
	   ) 	
	  -> true 
	  ;  remove_alarm => Event
	)
 ))
]) => mm_generic_event,

    put_slot([slot_tuple(primitive_call,global,set,optional,plog)]) => mm_inheritable_ms_event,

put_method([
   (retrieve_out_token(global, set, [], plog, [Token]) :- 
     	 (
	  message_recipient(LogicalEvent),
	  get_primitive_call([Tokens, Call]) => LogicalEvent,
	  call(Call),
	  member(Token,Tokens)
 ))
]) => mm_inheritable_ms_event,

put_method([
    (enable_event(global,[],[object],[],[Sender]) :-
	(
	 message_recipient(PEvent),
	 set_alarm => PEvent,	 
	 (my_instance_of(Sender,mm_production_rule) -> 
		(findall(Occ,(retrieve_out_token(Token) => PEvent,
		      	      Occ = occurrence_tuple(PEvent,positive(Token))),
		 	 AllOccs),
		 schedule_all([AllOccs]) => mm_production_rule_manager);	
	 	(findall(NextOcc,(retrieve_out_token(Token) => PEvent,
				  Occ = occurrence_tuple(PEvent,positive(Token)),
				  checking([Occ],NextOcc) => Sender),
			 AllOccs),
		 signal([AllOccs]) => mm_device_event))
 ))
]) => mm_inheritable_ms_event,


put_method([
 (replace_out_token(global,[],[plog,plog],[],[OldToken,NewToken]) :-
   (	
	message_recipient(Event),
	update_out_token([OldToken,NewToken]) => Event,
	(OldToken=token([],[],_) -> % Are all tokens empty?
		(( (get_activated_events(_) => Event;
		   get_de_activated_events(_) => Event;
		   get_activated_rules(_) => Event;
		   get_de_activated_rules(_) => Event;
	           NewToken=token([],[],_) )  % Are all tokens empty?
	               -> true 
	               ;  set_alarm => Event));

		 ( \+ NewToken=token([],[],_) 
	               -> true 
	               ;  
	
			  ( ( (get(CompositeEvent)    => mm_composite_event, 
	     		       get_composed_by(Event) => CompositeEvent,
	     		       get_activated_rules(_) => CompositeEvent);
	    			get_activated_rules(_) => Event;
	    			get_de_activated_rules(_) => Event;
	    			get_activated_events(_) => Event;
	    			get_de_activated_events(_) => Event
	   			) 	
	  			-> true 
	  			;  remove_alarm => Event
			  )
		)
	)
 ))
]) => mm_generic_event,



replace_method([
 (put_is_it_enabled(global,[],[string],[],[YN]) :-
      (	message_recipient(Rule),
	(YN == yes
	 -> (get_event(Event) => Rule,
	     put_activated_rules([Rule]) => Event,
	     (my_instance_of(Rule,mm_production_rule) ->
		 (get_anti_event(AEvent) => Rule,
		  put_de_activated_rules([Rule]) => AEvent);
		 true))
	 ;  true
	),
	is_it_enabled <-- YN, !
 ))
]) => mm_generic_rule,


replace_method([
 (update_is_it_enabled(global,[],[string,string],[],[Old,New]) :-
      (	message_recipient(Rule),
	Args = [Old,New],
	get_event(Event) => Rule,
	(	Args == [no,yes] -> put_activated_rules([Rule]) => Event
	  ;	Args == [yes,no] -> delete_activated_rules([Rule]) => Event
	),
	(my_instance_of(Rule,mm_production_rule) ->
		 (get_anti_event(AEvent) => Rule,
		  (Args == [no,yes] -> 
			put_de_activated_rules([Rule]) => AEvent;
	  		(Args == [yes,no] -> 
				delete_de_activated_rules([Rule]) => AEvent;
				true)));
		 true),
	is_it_enabled  <-- Old/New, !
 ))
]) => mm_generic_rule,

% The following are method re-definitions for the slot-accesing methods of
% the attribute 'activated_rules' of mm_generic_event
% Considerations for activated events has been added, too.
replace_method([
 (put_activated_rules(global,[],[mm_generic_rule],[],[Rule]) :-
   (	
	message_recipient(Event),
	% If the EVENT did not have any activated rule or event attached to it previously
	% then, SET_ALARM is sent for this event
	% The call for activated_events has been added by Nick for DEVICE
	( (get_activated_rules(_) => Event; get_activated_events(_) => Event)  
	  -> (activated_rules <-- Rule,
	      (my_instance_of(Rule,mm_production_rule) ->
			(findall(occurrence_tuple(Event,positive(Token)),			
				 retrieve_out_token(Token) =>  Event,
				 AllOccs),
			 schedule_all([AllOccs]) => mm_production_rule_manager);
			true))
	  ;  (activated_rules <-- Rule,
	      (my_instance_of(Rule,mm_production_rule) ->
			enable_event([Rule]) => Event;
			set_alarm => Event))			    
	),              
	(get_out_token(token(Activated,DeActivated,OutToken)) => Event ->
		((get_in_token(InToken) => Rule ->
		 	my_subset(InToken,OutToken);
			true),
		 update_out_token([token(Activated,DeActivated,OutToken),
			   	   token([Rule|Activated],DeActivated,OutToken)]) => Event);
		true), !
 ))
]) => mm_generic_event,


replace_method([
 (delete_activated_rules(global,[],[mm_generic_rule],[],[Rule]) :-
   (	
	message_recipient(Event),
	activated_rules <-- Rule/[],
	(my_instance_of(Rule,mm_production_rule) ->
		(findall(occurrence_tuple(Event,negative(Token)),			
			 retrieve_out_token(Token) =>  Event,
			 AllOccs),
		 schedule_all([AllOccs]) => mm_production_rule_manager);
		true),
	% If the EVENT does not have any activated rule or event left
	% then, REMOVE_ALARM is sent for this event
	% The call for activated_events has been added by Nick for DEVICE
	% The following 3 lines delete the deleted rule from the out_token slot.
	get_out_token(token(Activated,DeActivated,Token)) => Event,
	member(Rule,Activated), !,
	delete(Activated,Rule,Activated1),
	replace_out_token([token(Activated,DeActivated,Token),token(Activated1,DeActivated,Token)]) => Event
 ))
]) => mm_generic_event,

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
%	slot DE_ACTIVATED_RULES & method redefinition PUT_/DELETE_/UPDATE_ 
%
%	Definition of a new slot for the mm_two_input_event manager.
%	This was impossible to be done in the event manager creation as
%	the type of its contents are mm_production_rule, which has been  
%	recently defined
%
put_slot([slot_tuple(de_activated_rules,global,set,optional,mm_production_rule)])
      => mm_generic_event,


replace_method([
 (put_de_activated_rules(global,[],[mm_production_rule],[],[Rule]) :-
   (	message_recipient(Event),

	% If the EVENT did not have any de-activated rule or event attached to it previously
	% then, SET_ALARM is sent for this event
	%
	% Usually a complex event has both activated_rules and de_activated rules filled
	% with the same rules
	( (get_activated_rules(_) => Event; get_activated_events(_) => Event)  
	  -> true 
	  ;  set_alarm => Event
	),
	de_activated_rules <-- Rule,
	(get_out_token(token(Activated,DeActivated,OutToken)) => Event ->
		((get_in_token(InToken) => Rule ->
		 	my_subset(InToken,OutToken);
			true),
		 update_out_token([token(Activated,DeActivated,OutToken),
			   	   token(Activated,[Rule|DeActivated],OutToken)]) => Event);
		true), !

 ))
]) => mm_generic_event,


replace_method([
 (delete_de_activated_rules(global,[],[mm_production_rule],[],[Rule]) :-
   (	message_recipient(Event),
	de_activated_rules <-- Rule/[],
	% The following 3 lines delete the deleted rule from the out_token slot.
	get_out_token(token(Activated,DeActivated,Token)) => Event,
	member(Rule,DeActivated), !,
	delete(DeActivated,Rule,DeActivated1),
	replace_out_token([token(Activated,DeActivated,Token),token(Activated,DeActivated1,Token)]) => Event
 ))
]) => mm_generic_event.


% The 'detector_on' method is modified in order not to store many 
% times the same event alarm into ADAM (see file 'event_manager.pl' in EXACT) 
:- 
 replace_method([(detector_on(global,[],[string,string,sm_beh],[],
				  [When,ActiveMethod,Detector]) :-
     (	message_recipient(Event),

	% Translation of the 'when' value from the EXACT value to the ADAM value:
	%
	member(AdamWhen-When,[forward-before, backward-before_each, true-after,
			      fail-failure,   not_found-not_found]),
        new_name(ActiveMethod,StoreActiveMethod),
	Aux1 =.. [StoreActiveMethod,mm_rule_manager,AdamWhen-Detector-(Event)],
	(clause(Aux1,true) ->   % NICK! Here is my addition!
		true;
		store(Aux1)),
	(\+ mm_name(Event,_)
	 ->	(Aux2 =.. [StoreActiveMethod,rule_manager,AdamWhen-Detector-(Event)],
		 store(bang,_,Aux2))
	 ;	true)
    ))
]) => mm_ms_event.
