:- dynamic kb_module/2, pre_rule/3, rule/6.


load_kb(KBFile):-
	load_kb1(KBFile), !,
	(error_flag(1) 
		-> fail 
		 ; load_rules ).


load_rules:-
	retract(rule(ModName,ModPrior,RuleName,Control,RulePrior,RuleText)),
	  load_rule(ModName,ModPrior,RuleName,Control,RulePrior,RuleText),
	fail;true.



% MODULE PRIORITY HAS BEEN LEFT OUT
load_rule(ModName,_ModPrior,RuleName,Control,RulePrior,RuleText):-
	build_new_itemset([	module([ModName]),
				name([RuleName]),
				control([Control]),
				priority([RulePrior]),
				rule([RuleText]),
				is_it_enabled([yes])],ItemSet),				
	(new([_,ItemSet]) => mm_production_rule ->
		(clause('$spy_loader'(on),true) -> 
			(nonvar(RuleName) ->
				write_list(['Rule',RuleName,'inserted']);
				(write_list(['Rule:',RuleText,', inserted']), nl));
			true);
		(clause('$spy_loader'(on),true) -> 
			(nonvar(RuleName) ->
				write_list(['Rule',RuleName,'not inserted']);
				(write_list(['Rule:',RuleText,', not inserted']), nl));
			true)).


build_new_itemset([],[]).
build_new_itemset([Item|Items],Rest):-
	Item=..[_,[X]],
	var(X), !,
	build_new_itemset(Items,Rest).
build_new_itemset([Item|Items],[Item|Rest]):-
	build_new_itemset(Items,Rest).


load_kb1(FileName):-
	init,
	open(FileName,read,InputStream),
	repeat,
	  get0(InputStream,Char),
	  consume_char(Char,InputStream), 
	termination_conditions(InputStream), !,
	close(InputStream).
	
init:-
	retractall(rule(_,_,_,_,_,_)),
	retractall(error_flag(_)), assert(error_flag(0)),
	retractall(expected(_)),   assert(expected([start])), 
	assert(kb_module(_,_)).  % need it if there is a rule after start


% succeeds if EOS reached and no header-block is open
termination_conditions(InputStream):-
	at_end_of_stream(InputStream), !,
	expected(ExpectedHeaders),
	(ExpectedHeaders\==[]
		-> write_list(['SYNTAX ERROR: End of file reached while waiting for any of ',ExpectedHeaders,'.']),
		   rise_error_flag
		;  true	).
	
		   	   
%===== ignore comment =================
% when a % (ASCII=37) char is encountered then the rest on the
% line is considered to be a comment and is rejected
consume_char(37,InputStream):-
	!,
	repeat,
	  get0(InputStream,Char),
	end_of_comment(Char,InputStream), !.
%===== get header text =================
% when a # (ASCII=35) char is encountered then do
consume_char(35,InputStream):- 
	!,
	read(InputStream,Term),     	% read the header term
	Term=..[Functor|_],		% decompose it
	expected(ExpectedFunctors), 	% get the expected headers
	(ExpectedFunctors==[]		% should be [] if Functor=end 
		-> skip_rest_stream(InputStream)
		; (member(Functor,ExpectedFunctors) 	% if is expected
			-> update_expected(Functor),
		   	   !, 
			   parse_header(Term) 
			;  write_list(['SYNTAX ERROR: Any of ',ExpectedFunctors,' expected but ',Functor,' found.']),
			   rise_error_flag,
		   	   skip_rest_stream(InputStream) ) ).
%===== get rule text ===================
% when a " (ASCII=34) char is encountered then do
consume_char(34,InputStream):-    
	pre_rule(Name,Control,Priority), !, 	% get the stored (by parse_header/1) attributes
	read_rule(InputStream,T),		% read the body text
	strip_string(T,Stripped),		% strip all trailing spaces, tabs and CRs
	name(BodyRule,Stripped),		% convert to atom
	kb_module(ModuleName,ModulePriority), !,% get the module's attributes
	GoodRule=rule(ModuleName,ModulePriority,Name,Control,Priority,BodyRule),
	%GoodRule=rule(ModuleName,ModulePriority,Name,Control,Priority,Stripped),
	assert(GoodRule).
	

%===== all other crap ===================	
% ignores all other text
consume_char(_,_). 	

	
%=================================================================
% the header found was a module/2
parse_header(module(Name,Priority)):- !,
	retractall(kb_module(_,_)),
	assert(kb_module(Name,Priority)).
% the header found was a rule/3

parse_header(rule(Name,Control,Priority)):- !,
	retractall(pre_rule(_,_,_)),
	(Control==control
		-> assert(pre_rule(Name,yes,Priority))
		 ; assert(pre_rule(Name,Control,Priority)) ).
% the header found was a endmodule 
% NOTE: This case allows us to put rules not belonging to
% any module in any place we like :-)

parse_header(endmodule):- !,
	retractall(kb_module(_,_)),
	assert(kb_module(_,_)).
% the header found was any of: start, end, endrule

parse_header(_,[]).

%==========================================================================
% this one reads the bodytext of a rule

% (+,-) returns the text found between " "
read_rule(InputStream,[Char1|T]):-
	\+ at_end_of_stream(InputStream),
	get0(InputStream,Char),
	(Char=37	% if comment (ASCII=37) found during rule reading
		-> consume_char(37,InputStream), % reject comment
		   Char1=32, 			 % replace % with space (ASCII=32)
		   read_rule(InputStream,T)	 % read rest rule
		 ; (Char=34  		   % if " (ASCII=34) found then the rule has been parsed
			-> Char1=32, T=[]  % replace " with space (ASCII=32) and fix the tail
			 ; Char1=Char, read_rule(InputStream,T))).   % keep Char and read rest rule
% case of EOS during bodyrule reading
read_rule(InputStream,[]):-
	at_end_of_stream(InputStream),
	!,
	write('KB ERROR: End_Of_Stream reached during rule reading'), nl,
	rise_error_flag.
read_rule(_,[]).


%==========================================================================
	
% (+ASCIIList, -StrippedASCIIList)

strip_string(String,Stripped):-
	strip_non_printable(String,String1),	  % replaces tabs(9) and CR?(10) with spaces (32)
	strip_trailing_spaces(String1,Stripped).  % rejects trailing spaces

strip_non_printable([],[]).
strip_non_printable([X|RestIn],[32|RestOut]):-		% tab or CR
	member(X,[9,10]),
	!,
	strip_non_printable(RestIn,RestOut).
strip_non_printable([X|RestIn],[X|RestOut]):-
	strip_non_printable(RestIn,RestOut).

strip_trailing_spaces([],[]).
strip_trailing_spaces([32,32|RestIn],Out):-		%trailing spaces
	!,
	strip_trailing_spaces([32|RestIn],Out).
strip_trailing_spaces([X|RestIn],[X|RestOut]):-
	strip_trailing_spaces(RestIn,RestOut).

%========================================================

% updates the expected functors
update_expected(Functor):- 
	retractall(expected(_)),
	functor_pair(Functor,NewExpectedFunctors), !,
	assert(expected(NewExpectedFunctors)).

% (FunctorFound, FunctorsExpected)
functor_pair(module,[rule]).
functor_pair(rule,[endrule]).
functor_pair(endrule,[rule,module,endmodule,end]).
functor_pair(start,[rule,module]).
functor_pair(endmodule,[rule,module,end]).
functor_pair(end,[]).


% succeeds if EOS is reached or ASCII=10 is encoundtered
end_of_comment(_,InputStream):-	at_end_of_stream(InputStream), !.
end_of_comment(10,_).


% rejects the rest of the stream
skip_rest_stream(InputStream):-
	repeat,
	  get0(InputStream,_),
	at_end_of_stream(InputStream), !.


rise_error_flag:-
	retractall(error_flag(_)),
	assert(error_flag(1)).

%========================================================
