/* optimizer.pl */

optimize([CondIn,Action],[CondOut,Action]) :- 
	(append(Cond,[prolog(Goal)],CondIn) -> 
		true;
		Cond = CondIn),
	opt_intra(Cond,ScoredCondNextPos,ScoredCondNextNeg),
	opt_inter(ScoredCondNextPos,CondPos),
	opt_inter(ScoredCondNextNeg,CondNeg),
	append(CondPos,CondNeg,Cond1),
	(Cond == CondIn ->
		CondOut	= Cond1;
		append(Cond1,[prolog(Goal)],CondOut)).

opt_intra(Cond,ScoredCondNextPos,ScoredCondNextNeg) :- 
	opt_intra(Cond,[],ScoredCondNextPos,[],ScoredCondNextNeg).


opt_intra([],CondPos,CondPos,CondNeg,CondNeg) :- !.
opt_intra([nested_neg(IOPs)|RestCond],TempCondPos,CondPos,TempCondNeg,CondNeg) :- 
	opt_intra(IOPs,ScoredIOPPos,ScoredIOPNeg),
	opt_inter(ScoredIOPPos,IOPPos),
	opt_inter(ScoredIOPNeg,IOPNeg),
	total_scores(ScoredIOPPos,NoTestsP,ScoreTestsP),
	total_scores(ScoredIOPNeg,NoTestsN,ScoreTestsN),
	NoTests is NoTestsP + NoTestsN,
	ScoreTests is ScoreTestsP + ScoreTestsN,
	append(IOPPos,IOPNeg,IOPOut),
	opt_intra(RestCond,TempCondPos,CondPos,[[NoTests,ScoreTests]-nested_neg(IOPOut)|TempCondNeg],CondNeg).
opt_intra([neg([OID_Var,Class,AttribPatternsIn])|RestCond],TempCondPos,CondPos,TempCondNeg,CondNeg) :- 
	score_tests(AttribPatternsIn,AttribPatternsOut,NoTests,ScoreTests),
	opt_intra(RestCond,TempCondPos,CondPos,
			[[NoTests,ScoreTests]-neg([OID_Var,Class,AttribPatternsOut])|TempCondNeg],CondNeg).
opt_intra([pos([OID_Var,Class,AttribPatternsIn])|RestCond],TempCondPos,CondPos,TempCondNeg,CondNeg) :- 
	score_tests(AttribPatternsIn,AttribPatternsOut,NoTests,ScoreTests),
	opt_intra(RestCond,[[NoTests,ScoreTests]-pos([OID_Var,Class,AttribPatternsOut])|TempCondPos],CondPos,
			TempCondNeg,CondNeg).

opt_inter(ScoredConds,Conds) :- 
	keysort(ScoredConds,SortedScoredConds),
	reverse(SortedScoredConds,ReversedSortedScoredConds),
	findall(Cond,member(_-Cond,ReversedSortedScoredConds),Conds).


score_tests(AttribPatternsIn,AttribPatternsOut,NoTests,ScoreTests) :- 
	score_tests(AttribPatternsIn,TestPatterns1,TestPatterns2,JoinPatterns,DefPatterns,0,NoTests,0,ScoreTests),
	append(JoinPatterns,DefPatterns,AttribPatterns1),
	append(TestPatterns2,AttribPatterns1,AttribPatterns2),
	append(TestPatterns1,AttribPatterns2,AttribPatternsOut).

score_tests([],[],[],[],[],NoTests,NoTests,ScoreTests,ScoreTests) :- !.
score_tests([def(Attr,var(Var))|RestPatterns],TestPatterns1,TestPatterns2,JoinPatterns,
		[def(Attr,var(Var))|DefPatternsTail],
		TempNoTests,NoTests,TempScoreTests,ScoreTests) :- !,
	score_tests(RestPatterns,TestPatterns1,TestPatterns2,JoinPatterns,
			DefPatternsTail,TempNoTests,NoTests,TempScoreTests,ScoreTests).
score_tests([defcomp(Attr,var(Var1),pred(R,var(Var2)))|RestPatterns],TestPatterns1,TestPatterns2,
		[defcomp(Attr,var(Var1),pred(R,var(Var2)))|JoinPatternsTail],DefPatterns,
		TempNoTests,NoTests,TempScoreTests,ScoreTests) :- !,
	score_tests(RestPatterns,TestPatterns1,TestPatterns2,JoinPatternsTail,
			DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests).
score_tests([comp(Attr,pred(R,var(Var)))|RestPatterns],TestPatterns1,TestPatterns2,
		[comp(Attr,pred(R,var(Var)))|JoinPatternsTail],DefPatterns,
		TempNoTests,NoTests,TempScoreTests,ScoreTests) :- !,
	score_tests(RestPatterns,TestPatterns1,TestPatterns2,JoinPatternsTail,
			DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests).
score_tests([comp(Attr,Preds)|RestPatterns],[comp(Attr,Preds)|TestPatterns1Tail],TestPatterns2,
		JoinPatterns,DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests) :- 
	score_predicates(Preds,Score),
	Score >= 100, !,
	NextNoTests is TempNoTests + 1,
	NextScoreTests is TempScoreTests + Score,
	score_tests(RestPatterns,TestPatterns1Tail,TestPatterns2,JoinPatterns,
			DefPatterns,NextNoTests,NoTests,NextScoreTests,ScoreTests).
score_tests([comp(Attr,Preds)|RestPatterns],TestPatterns1,[comp(Attr,Preds)|TestPatterns2Tail],
		JoinPatterns,DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests) :- !,
	score_predicates(Preds,Score),
	NextNoTests is TempNoTests + 1,
	NextScoreTests is TempScoreTests + Score,
	score_tests(RestPatterns,TestPatterns1,TestPatterns2Tail,JoinPatterns,
			DefPatterns,NextNoTests,NoTests,NextScoreTests,ScoreTests).
score_tests([defcomp(Attr,var(Var),Preds)|RestPatterns],
		[defcomp(Attr,var(Var),Preds)|TestPatterns1Tail],TestPatterns2,
		JoinPatterns,DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests) :- 
	score_predicates(Preds,Score),
	Score >= 100, !,
	NextNoTests is TempNoTests + 1,
	NextScoreTests is TempScoreTests + Score,
	score_tests(RestPatterns,TestPatterns1Tail,TestPatterns2,JoinPatterns,
			DefPatterns,NextNoTests,NoTests,NextScoreTests,ScoreTests).
score_tests([defcomp(Attr,var(Var),Preds)|RestPatterns],TestPatterns1,
		[defcomp(Attr,var(Var),Preds)|TestPatterns2Tail],
		JoinPatterns,DefPatterns,TempNoTests,NoTests,TempScoreTests,ScoreTests) :- 
	score_predicates(Preds,Score),
	NextNoTests is TempNoTests + 1,
	NextScoreTests is TempScoreTests + Score,
	score_tests(RestPatterns,TestPatterns1,TestPatterns2Tail,JoinPatterns,
			DefPatterns,NextNoTests,NoTests,NextScoreTests,ScoreTests).


score_predicates(single([pred(Op,_Const)]),Score) :-
	score_operator(Op,Score).
score_predicates(and(Preds),Score) :- 
	score_all_operators(Preds,Scores),
	find_max(Scores,Score).
score_predicates(or(Preds),Score) :- 
	score_all_operators(Preds,Scores),
	find_min(Scores,Score).

score_all_operators([],[]) :- !.
score_all_operators([pred(Op,_Const)|RestPreds],[Score|RestScores]) :- 
	score_operator(Op,Score),
	score_all_operators(RestPreds,RestScores).

score_operator('=',100) :- !.
score_operator(_,1).


total_scores(ScoredConds,NoTests,ScoreTests) :- 
	total_scores(ScoredConds,0,NoTests,0,ScoreTests).

total_scores([],NoTests,NoTests,ScoreTests,ScoreTests) :- !.
total_scores([[NoTests1,ScoreTests1]-_|RestScoredConds],TempNoTests,NoTests,TempScoreTests,ScoreTests) :- 
	NextNoTests is TempNoTests + NoTests1,
	NextScoreTests is TempScoreTests + ScoreTests1,
	total_scores(RestScoredConds,NextNoTests,NoTests,NextScoreTests,ScoreTests).

