/* precompiler.pl */

precompile(Vars,[Cond1,Action],[Cond2,Action]) :- 
	% These lines create multiple patterns to cater for complex chained attributes
	de_chain(Vars,Cond1,Lists),
	append_lists(Lists,Cond2),
	reset_var_gen.


% These lines create multiple patterns to cater for logically connected predicates in attribute tests


de_chain(_,[],[]) :- !.
de_chain(Vars,[H|T],[H1|T1]) :- 
	de_chain_one(Vars,H,H1),
	de_chain(Vars,T,T1).

de_chain_one(Vars,pos([V,Class,APatterns]),[pos([V,Class,NewAPatterns])|New_IntraPatterns]) :- !,
	de_chain_attributes(Vars,[],Class,APatterns,NewAPatterns,New_IntraPatterns).
de_chain_one(Vars,nested_neg(IOPs),[nested_neg(ExitIOPs)]) :- !,
	de_chain(Vars,IOPs,Lists),
	append_lists(Lists,ExitIOPs).	
de_chain_one(Vars,neg([V,Class,APatterns]),Return) :-
	de_chain_attributes(Vars,[],Class,APatterns,NewAPatterns,New_IntraPatterns),
	(New_IntraPatterns == [] -> 
	    Return = [neg([V,Class,NewAPatterns])];
	    Return = [nested_neg([pos([V,Class,NewAPatterns])|New_IntraPatterns])]).
de_chain_one(_,prolog(Goal),[prolog(Goal)]).

de_chain_attributes(_,_,_,[],[],[]) :- !.
de_chain_attributes(Vars,InsideVars,Class,[APattern|RestAPatterns],
	                 FinalNewAPatterns,All_New_IntraPatterns) :- 
	de_chain_one_attribute(Vars,InsideVars,Class,APattern,New_APattern,New_IntraPatterns),
	(New_IntraPatterns == [] -> 
		NewInsideVars = InsideVars;
		(member(New_APattern,InsideVars) -> 
			NewInsideVars = InsideVars;
			NewInsideVars = [New_APattern|InsideVars])),
	de_chain_attributes(Vars,NewInsideVars,Class,RestAPatterns,RestNewAPatterns,RestNew_IntraPatterns),
	concat_patterns(New_IntraPatterns,RestNew_IntraPatterns,All_New_IntraPatterns),
	(member(New_APattern,RestNewAPatterns) -> 
		FinalNewAPatterns = RestNewAPatterns;
		FinalNewAPatterns = [New_APattern|RestNewAPatterns]).

de_chain_one_attribute(_,_,_Class,APattern,NewPattern,[]) :- 
	arg(1,APattern,[Attribute]), !,
	my_setarg(1,APattern,Attribute,NewPattern).
de_chain_one_attribute(Vars,InsideVars,Class,APattern,def(Attribute,var(Var)),New_IntraPatterns) :- 
	(Var = unique(V) -> 
		true; 
		V = Var),
	arg(1,APattern,ListofAttributes),
	reverse(ListofAttributes,[Attribute|ChainedAtts]),
	(member(def(Attribute,var(V)),InsideVars) -> 
		true;
		gen_new_var(Vars,V)),
	find_next_class(Class,Attribute,NextClass),
	de_chain_one_chain(Vars,NextClass,APattern,V,ChainedAtts,New_IntraPatterns).

de_chain_one_chain(_Vars,Class,OriginalPattern,PreviousVar,[LastAttr],
	                [pos([PreviousVar,Class,[NewPattern]])]) :- !,
	my_setarg(1,OriginalPattern,LastAttr,NewPattern).
de_chain_one_chain(Vars,Class,OPattern,PrevVar,[FirstAttr|RestAttrs],
	                [pos([PrevVar,Class,[def(FirstAttr,var(Var))]])|RestIPs]) :- 
	(Var = unique(V) -> 
		true; 
		V = Var),
	gen_new_var(Vars,V),
	find_next_class(Class,FirstAttr,_NextClass),
	de_chain_one_chain(Vars,Class,OPattern,V,RestAttrs,RestIPs).

% This predicate 'merges' the different intra-object patterns that are produced by de-chaining
% two or more chained attributes with the same target class, into one intra-object pattern
% e.g. name.manager and salary.manager
concat_patterns([],Patterns2,Patterns2) :- !.
concat_patterns([P1|RPatterns1],Patterns2,[P3|RPatterns3]) :- 
	P1 = pos([Var,Class,AttrPatterns1]),
	delete_one_fail(Patterns2,pos([Var,Class,AttrPatterns2]),NextPatterns2), !,
	append(AttrPatterns1,AttrPatterns2,AttrPatterns3),
	P3 = pos([Var,Class,AttrPatterns3]),
	concat_patterns(RPatterns1,NextPatterns2,RPatterns3).
concat_patterns([P1|RPatterns1],Patterns2,[P1|RPatterns3]) :- 
	concat_patterns(RPatterns1,Patterns2,RPatterns3).
	

find_next_class(Class,Attr,Next) :- 
	path(Class,Supers),
	find_next_class_aux([Class|Supers],Attr,Next).

find_next_class_aux([Class|_],Attr,Next) :- 
	get_slot_desc(slot_desc_tuple(Attr,_,_,_,_,Next)) => Class, !.
find_next_class_aux([_|Supers],Attr,Next) :- 
	find_next_class_aux(Supers,Attr,Next).



